from __future__ import annotations

import pytest

from textual.color import Color
from textual.css.errors import UnresolvedVariableError
from textual.css.parse import substitute_references
from textual.css.scalar import Scalar, Unit
from textual.css.stylesheet import Stylesheet, StylesheetParseError
from textual.css.tokenize import tokenize
from textual.css.tokenizer import ReferencedBy, Token, TokenError
from textual.css.transition import Transition
from textual.geometry import Spacing
from textual.layouts.vertical import VerticalLayout


class TestVariableReferenceSubstitution:
    def test_simple_reference(self):
        css = "$x: 1; #some-widget{border: $x;}"
        variables = substitute_references(tokenize(css, ("", "")))
        assert list(variables) == [
            Token(
                name="variable_name",
                value="$x:",
                read_from=("", ""),
                code=css,
                location=(0, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 3),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="1",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=None,
            ),
            Token(
                name="variable_value_end",
                value=";",
                read_from=("", ""),
                code=css,
                location=(0, 5),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 6),
                referenced_by=None,
            ),
            Token(
                name="selector_start_id",
                value="#some-widget",
                read_from=("", ""),
                code=css,
                location=(0, 7),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_start",
                value="{",
                read_from=("", ""),
                code=css,
                location=(0, 19),
                referenced_by=None,
            ),
            Token(
                name="declaration_name",
                value="border:",
                read_from=("", ""),
                code=css,
                location=(0, 20),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 27),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="1",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=ReferencedBy(
                    name="x", location=(0, 28), length=2, code=css
                ),
            ),
            Token(
                name="declaration_end",
                value=";",
                read_from=("", ""),
                code=css,
                location=(0, 30),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_end",
                value="}",
                read_from=("", ""),
                code=css,
                location=(0, 31),
                referenced_by=None,
            ),
        ]

    def test_simple_reference_no_whitespace(self):
        css = "$x:1; #some-widget{border: $x;}"
        variables = substitute_references(tokenize(css, ("", "")))
        assert list(variables) == [
            Token(
                name="variable_name",
                value="$x:",
                read_from=("", ""),
                code=css,
                location=(0, 0),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="1",
                read_from=("", ""),
                code=css,
                location=(0, 3),
                referenced_by=None,
            ),
            Token(
                name="variable_value_end",
                value=";",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 5),
                referenced_by=None,
            ),
            Token(
                name="selector_start_id",
                value="#some-widget",
                read_from=("", ""),
                code=css,
                location=(0, 6),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_start",
                value="{",
                read_from=("", ""),
                code=css,
                location=(0, 18),
                referenced_by=None,
            ),
            Token(
                name="declaration_name",
                value="border:",
                read_from=("", ""),
                code=css,
                location=(0, 19),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 26),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="1",
                read_from=("", ""),
                code=css,
                location=(0, 3),
                referenced_by=ReferencedBy(
                    name="x", location=(0, 27), length=2, code=css
                ),
            ),
            Token(
                name="declaration_end",
                value=";",
                read_from=("", ""),
                code=css,
                location=(0, 29),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_end",
                value="}",
                read_from=("", ""),
                code=css,
                location=(0, 30),
                referenced_by=None,
            ),
        ]

    def test_undefined_variable(self):
        css = ".thing { border: $not-defined; }"
        with pytest.raises(UnresolvedVariableError):
            list(substitute_references(tokenize(css, ("", ""))))

    def test_empty_variable(self):
        css = "$x:\n* { background:$x; }"
        result = list(substitute_references(tokenize(css, ("", ""))))
        assert [(t.name, t.value) for t in result] == [
            ("variable_name", "$x:"),
            ("variable_value_end", "\n"),
            ("selector_start_universal", "*"),
            ("whitespace", " "),
            ("declaration_set_start", "{"),
            ("whitespace", " "),
            ("declaration_name", "background:"),
            ("declaration_end", ";"),
            ("whitespace", " "),
            ("declaration_set_end", "}"),
        ]

    def test_transitive_reference(self):
        css = "$x: 1\n$y: $x\n.thing { border: $y }"
        assert list(substitute_references(tokenize(css, ("", "")))) == [
            Token(
                name="variable_name",
                value="$x:",
                read_from=("", ""),
                code=css,
                location=(0, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 3),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="1",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=None,
            ),
            Token(
                name="variable_value_end",
                value="\n",
                read_from=("", ""),
                code=css,
                location=(0, 5),
                referenced_by=None,
            ),
            Token(
                name="variable_name",
                value="$y:",
                read_from=("", ""),
                code=css,
                location=(1, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 3),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="1",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=ReferencedBy(
                    name="x", location=(1, 4), length=2, code=css
                ),
            ),
            Token(
                name="variable_value_end",
                value="\n",
                read_from=("", ""),
                code=css,
                location=(1, 6),
                referenced_by=None,
            ),
            Token(
                name="selector_start_class",
                value=".thing",
                read_from=("", ""),
                code=css,
                location=(2, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(2, 6),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_start",
                value="{",
                read_from=("", ""),
                code=css,
                location=(2, 7),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(2, 8),
                referenced_by=None,
            ),
            Token(
                name="declaration_name",
                value="border:",
                read_from=("", ""),
                code=css,
                location=(2, 9),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(2, 16),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="1",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=ReferencedBy(
                    name="y", location=(2, 17), length=2, code=css
                ),
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(2, 19),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_end",
                value="}",
                read_from=("", ""),
                code=css,
                location=(2, 20),
                referenced_by=None,
            ),
        ]

    def test_multi_value_variable(self):
        css = "$x: 2 4\n$y: 6 $x 2\n.thing { border: $y }"
        assert list(substitute_references(tokenize(css, ("", "")))) == [
            Token(
                name="variable_name",
                value="$x:",
                read_from=("", ""),
                code=css,
                location=(0, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 3),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="2",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 5),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="4",
                read_from=("", ""),
                code=css,
                location=(0, 6),
                referenced_by=None,
            ),
            Token(
                name="variable_value_end",
                value="\n",
                read_from=("", ""),
                code=css,
                location=(0, 7),
                referenced_by=None,
            ),
            Token(
                name="variable_name",
                value="$y:",
                read_from=("", ""),
                code=css,
                location=(1, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 3),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="6",
                read_from=("", ""),
                code=css,
                location=(1, 4),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 5),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="2",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=ReferencedBy(
                    name="x", location=(1, 6), length=2, code=css
                ),
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 5),
                referenced_by=ReferencedBy(
                    name="x", location=(1, 6), length=2, code=css
                ),
            ),
            Token(
                name="number",
                value="4",
                read_from=("", ""),
                code=css,
                location=(0, 6),
                referenced_by=ReferencedBy(
                    name="x", location=(1, 6), length=2, code=css
                ),
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 8),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="2",
                read_from=("", ""),
                code=css,
                location=(1, 9),
                referenced_by=None,
            ),
            Token(
                name="variable_value_end",
                value="\n",
                read_from=("", ""),
                code=css,
                location=(1, 10),
                referenced_by=None,
            ),
            Token(
                name="selector_start_class",
                value=".thing",
                read_from=("", ""),
                code=css,
                location=(2, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(2, 6),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_start",
                value="{",
                read_from=("", ""),
                code=css,
                location=(2, 7),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(2, 8),
                referenced_by=None,
            ),
            Token(
                name="declaration_name",
                value="border:",
                read_from=("", ""),
                code=css,
                location=(2, 9),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(2, 16),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="6",
                read_from=("", ""),
                code=css,
                location=(1, 4),
                referenced_by=ReferencedBy(
                    name="y", location=(2, 17), length=2, code=css
                ),
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 5),
                referenced_by=ReferencedBy(
                    name="y", location=(2, 17), length=2, code=css
                ),
            ),
            Token(
                name="number",
                value="2",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=ReferencedBy(
                    name="y", location=(2, 17), length=2, code=css
                ),
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 5),
                referenced_by=ReferencedBy(
                    name="y", location=(2, 17), length=2, code=css
                ),
            ),
            Token(
                name="number",
                value="4",
                read_from=("", ""),
                code=css,
                location=(0, 6),
                referenced_by=ReferencedBy(
                    name="y", location=(2, 17), length=2, code=css
                ),
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 8),
                referenced_by=ReferencedBy(
                    name="y", location=(2, 17), length=2, code=css
                ),
            ),
            Token(
                name="number",
                value="2",
                read_from=("", ""),
                code=css,
                location=(1, 9),
                referenced_by=ReferencedBy(
                    name="y", location=(2, 17), length=2, code=css
                ),
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(2, 19),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_end",
                value="}",
                read_from=("", ""),
                code=css,
                location=(2, 20),
                referenced_by=None,
            ),
        ]

    def test_variable_used_inside_property_value(self):
        css = "$x: red\n.thing { border: on $x; }"
        assert list(substitute_references(tokenize(css, ("", "")))) == [
            Token(
                name="variable_name",
                value="$x:",
                read_from=("", ""),
                code=css,
                location=(0, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 3),
                referenced_by=None,
            ),
            Token(
                name="token",
                value="red",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=None,
            ),
            Token(
                name="variable_value_end",
                value="\n",
                read_from=("", ""),
                code=css,
                location=(0, 7),
                referenced_by=None,
            ),
            Token(
                name="selector_start_class",
                value=".thing",
                read_from=("", ""),
                code=css,
                location=(1, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 6),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_start",
                value="{",
                read_from=("", ""),
                code=css,
                location=(1, 7),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 8),
                referenced_by=None,
            ),
            Token(
                name="declaration_name",
                value="border:",
                read_from=("", ""),
                code=css,
                location=(1, 9),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 16),
                referenced_by=None,
            ),
            Token(
                name="token",
                value="on",
                read_from=("", ""),
                code=css,
                location=(1, 17),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 19),
                referenced_by=None,
            ),
            Token(
                name="token",
                value="red",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=ReferencedBy(
                    name="x", location=(1, 20), length=2, code=css
                ),
            ),
            Token(
                name="declaration_end",
                value=";",
                read_from=("", ""),
                code=css,
                location=(1, 22),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(1, 23),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_end",
                value="}",
                read_from=("", ""),
                code=css,
                location=(1, 24),
                referenced_by=None,
            ),
        ]

    def test_variable_definition_eof(self):
        css = "$x: 1"
        assert list(substitute_references(tokenize(css, ("", "")))) == [
            Token(
                name="variable_name",
                value="$x:",
                read_from=("", ""),
                code=css,
                location=(0, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 3),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="1",
                read_from=("", ""),
                code=css,
                location=(0, 4),
                referenced_by=None,
            ),
        ]

    def test_variable_reference_whitespace_trimming(self):
        css = "$x:    123;.thing{border: $x}"
        assert list(substitute_references(tokenize(css, ("", "")))) == [
            Token(
                name="variable_name",
                value="$x:",
                read_from=("", ""),
                code=css,
                location=(0, 0),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value="    ",
                read_from=("", ""),
                code=css,
                location=(0, 3),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="123",
                read_from=("", ""),
                code=css,
                location=(0, 7),
                referenced_by=None,
            ),
            Token(
                name="variable_value_end",
                value=";",
                read_from=("", ""),
                code=css,
                location=(0, 10),
                referenced_by=None,
            ),
            Token(
                name="selector_start_class",
                value=".thing",
                read_from=("", ""),
                code=css,
                location=(0, 11),
                referenced_by=None,
            ),
            Token(
                name="declaration_set_start",
                value="{",
                read_from=("", ""),
                code=css,
                location=(0, 17),
                referenced_by=None,
            ),
            Token(
                name="declaration_name",
                value="border:",
                read_from=("", ""),
                code=css,
                location=(0, 18),
                referenced_by=None,
            ),
            Token(
                name="whitespace",
                value=" ",
                read_from=("", ""),
                code=css,
                location=(0, 25),
                referenced_by=None,
            ),
            Token(
                name="number",
                value="123",
                read_from=("", ""),
                code=css,
                location=(0, 7),
                referenced_by=ReferencedBy(
                    name="x", location=(0, 26), length=2, code=css
                ),
            ),
            Token(
                name="declaration_set_end",
                value="}",
                read_from=("", ""),
                code=css,
                location=(0, 28),
                referenced_by=None,
            ),
        ]


class TestParseLayout:
    def test_valid_layout_name(self):
        css = "#some-widget { layout: vertical; }"

        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles
        assert isinstance(styles.layout, VerticalLayout)

    def test_invalid_layout_name(self):
        css = "#some-widget { layout: invalidlayout; }"

        stylesheet = Stylesheet()
        with pytest.raises(StylesheetParseError) as ex:
            stylesheet.add_source(css)
            stylesheet.parse()

        assert ex.value.errors is not None


class TestParseText:
    def test_foreground(self):
        css = """#some-widget {
            color: green;
        }
        """
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles
        assert styles.color == Color.parse("green")

    def test_background(self):
        css = """#some-widget {
            background: red;
        }
        """
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles
        assert styles.background == Color.parse("red")


class TestParseColor:
    """More in-depth tests around parsing of CSS colors"""

    @pytest.mark.parametrize(
        "value,result",
        [
            ("rgb(1,255,50)", Color(1, 255, 50)),
            ("rgb( 1, 255,50 )", Color(1, 255, 50)),
            ("rgba( 1, 255,50,0.3 )", Color(1, 255, 50, 0.3)),
            ("rgba( 1, 255,50, 1.3 )", Color(1, 255, 50, 1.0)),
            ("hsl( 180, 50%, 50% )", Color(64, 191, 191)),
            ("hsl(180,50%,50%)", Color(64, 191, 191)),
            ("hsla(180,50%,50%,0.25)", Color(64, 191, 191, 0.25)),
            ("hsla( 180, 50% ,50%,0.25 )", Color(64, 191, 191, 0.25)),
            ("hsla( 180, 50% , 50% , 1.5 )", Color(64, 191, 191)),
        ],
    )
    def test_rgb_and_hsl(self, value, result):
        css = f""".box {{
          color: {value};
        }}
        """
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles
        assert styles.color == result


class TestParseOffset:
    @pytest.mark.parametrize(
        "offset_x, parsed_x, offset_y, parsed_y",
        [
            [
                "-5.5%",
                Scalar(-5.5, Unit.PERCENT, Unit.WIDTH),
                "-30%",
                Scalar(-30, Unit.PERCENT, Unit.HEIGHT),
            ],
            [
                "5%",
                Scalar(5, Unit.PERCENT, Unit.WIDTH),
                "40%",
                Scalar(40, Unit.PERCENT, Unit.HEIGHT),
            ],
            [
                "10",
                Scalar(10, Unit.CELLS, Unit.WIDTH),
                "40",
                Scalar(40, Unit.CELLS, Unit.HEIGHT),
            ],
        ],
    )
    def test_composite_rule(self, offset_x, parsed_x, offset_y, parsed_y):
        css = f"""#some-widget {{
            offset: {offset_x} {offset_y};
        }}
        """
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles

        assert len(stylesheet.rules) == 1
        assert stylesheet.rules[0].errors == []
        assert styles.offset.x == parsed_x
        assert styles.offset.y == parsed_y

    @pytest.mark.parametrize(
        "offset_x, parsed_x, offset_y, parsed_y",
        [
            [
                "-5.5%",
                Scalar(-5.5, Unit.PERCENT, Unit.WIDTH),
                "-30%",
                Scalar(-30, Unit.PERCENT, Unit.HEIGHT),
            ],
            [
                "5%",
                Scalar(5, Unit.PERCENT, Unit.WIDTH),
                "40%",
                Scalar(40, Unit.PERCENT, Unit.HEIGHT),
            ],
            [
                "-10",
                Scalar(-10, Unit.CELLS, Unit.WIDTH),
                "40",
                Scalar(40, Unit.CELLS, Unit.HEIGHT),
            ],
        ],
    )
    def test_separate_rules(self, offset_x, parsed_x, offset_y, parsed_y):
        css = f"""#some-widget {{
            offset-x: {offset_x};
            offset-y: {offset_y};
        }}
        """
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles

        assert len(stylesheet.rules) == 1
        assert stylesheet.rules[0].errors == []
        assert styles.offset.x == parsed_x
        assert styles.offset.y == parsed_y


class TestParseOverflow:
    def test_multiple_enum(self):
        css = "#some-widget { overflow: hidden auto; }"
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles

        assert len(stylesheet.rules) == 1
        assert styles.overflow_x == "hidden"
        assert styles.overflow_y == "auto"


class TestParseTransition:
    @pytest.mark.parametrize(
        "duration, parsed_duration",
        [
            ["5.57s", 5.57],
            ["0.5s", 0.5],
            ["1200ms", 1.2],
            ["0.5ms", 0.0005],
            ["20", 20.0],
            ["0.1", 0.1],
        ],
    )
    def test_various_duration_formats(self, duration, parsed_duration):
        easing = "in_out_cubic"
        transition_property = "offset"
        delay = duration
        css = f"""#some-widget {{
            transition: {transition_property} {duration} {easing} {delay};
        }}
        """
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles

        assert len(stylesheet.rules) == 1
        assert stylesheet.rules[0].errors == []
        assert styles.transitions == {
            "offset": Transition(
                duration=parsed_duration, easing=easing, delay=parsed_duration
            )
        }

    def test_no_delay_specified(self):
        css = f"#some-widget {{ transition: offset-x 1 in_out_cubic; }}"
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        styles = stylesheet.rules[0].styles

        assert stylesheet.rules[0].errors == []
        assert styles.transitions == {
            "offset-x": Transition(duration=1, easing="in_out_cubic", delay=0)
        }

    def test_unknown_easing_function(self):
        invalid_func_name = "invalid_easing_function"
        css = f"#some-widget {{ transition: offset 1 {invalid_func_name} 1; }}"

        stylesheet = Stylesheet()
        with pytest.raises(StylesheetParseError) as ex:
            stylesheet.add_source(css)
            stylesheet.parse()

        rules = stylesheet._parse_rules(css, "foo")
        stylesheet_errors = rules[0].errors

        assert len(stylesheet_errors) == 1
        assert stylesheet_errors[0][0].value == invalid_func_name
        assert ex.value.errors is not None


class TestParseOpacity:
    @pytest.mark.parametrize(
        "css_value, styles_value",
        [
            ["-0.2", 0.0],
            ["0.4", 0.4],
            ["1.3", 1.0],
            ["-20%", 0.0],
            ["25%", 0.25],
            ["128%", 1.0],
        ],
    )
    def test_opacity_to_styles(self, css_value, styles_value):
        css = f"#some-widget {{ text-opacity: {css_value} }}"
        stylesheet = Stylesheet()
        stylesheet.add_source(css)

        assert stylesheet.rules[0].styles.text_opacity == styles_value
        assert not stylesheet.rules[0].errors

    def test_opacity_invalid_value(self):
        css = "#some-widget { text-opacity: 123x }"
        stylesheet = Stylesheet()

        with pytest.raises(StylesheetParseError):
            stylesheet.add_source(css)
            stylesheet.parse()
        rules = stylesheet._parse_rules(css, "foo")
        assert rules[0].errors


class TestParseMargin:
    def test_margin_partial(self):
        css = "#foo {margin: 1; margin-top: 2; margin-right: 3; margin-bottom: -1;}"
        stylesheet = Stylesheet()
        stylesheet.add_source(css)
        assert stylesheet.rules[0].styles.margin == Spacing(2, 3, -1, 1)


class TestParsePadding:
    def test_padding_partial(self):
        css = "#foo {padding: 1; padding-top: 2; padding-right: 3; padding-bottom: -1;}"
        stylesheet = Stylesheet()
        stylesheet.add_source(css)
        assert stylesheet.rules[0].styles.padding == Spacing(2, 3, -1, 1)


class TestParseTextAlign:
    @pytest.mark.parametrize(
        "valid_align", ["left", "start", "center", "right", "end", "justify"]
    )
    def test_text_align(self, valid_align):
        css = f"#foo {{ text-align: {valid_align} }}"
        stylesheet = Stylesheet()
        stylesheet.add_source(css)
        assert stylesheet.rules[0].styles.text_align == valid_align

    def test_text_align_invalid(self):
        css = "#foo { text-align: invalid-value; }"
        stylesheet = Stylesheet()
        with pytest.raises(StylesheetParseError):
            stylesheet.add_source(css)
            stylesheet.parse()
        rules = stylesheet._parse_rules(css, "foo")
        assert rules[0].errors


class TestTypeNames:
    def test_type_no_number(self):
        stylesheet = Stylesheet()
        stylesheet.add_source("TestType {}")
        assert len(stylesheet.rules) == 1

    def test_type_with_number(self):
        stylesheet = Stylesheet()
        stylesheet.add_source("TestType1 {}")
        assert len(stylesheet.rules) == 1

    def test_type_starts_with_number(self):
        stylesheet = Stylesheet()
        stylesheet.add_source("1TestType {}")
        with pytest.raises(TokenError):
            stylesheet.parse()

    def test_combined_type_no_number(self):
        for separator in " >,":
            stylesheet = Stylesheet()
            stylesheet.add_source(f"StartType {separator} TestType {{}}")
            assert len(stylesheet.rules) == 1

    def test_combined_type_with_number(self):
        for separator in " >,":
            stylesheet = Stylesheet()
            stylesheet.add_source(f"StartType {separator} TestType1 {{}}")
            assert len(stylesheet.rules) == 1

    def test_combined_type_starts_with_number(self):
        for separator in " >,":
            stylesheet = Stylesheet()
            stylesheet.add_source(f"StartType {separator} 1TestType {{}}")
            with pytest.raises(TokenError):
                stylesheet.parse()


def test_parse_bad_pseudo_selector():
    """Check unknown selector raises a token error."""

    bad_selector = """\
Widget:foo{
    border: red;
}
    """

    stylesheet = Stylesheet()
    stylesheet.add_source(bad_selector, None)

    with pytest.raises(TokenError) as error:
        stylesheet.parse()

    assert error.value.start == (1, 7)


def test_parse_bad_pseudo_selector_with_suggestion():
    """Check unknown pseudo selector raises token error with correct position."""

    bad_selector = """
Widget:blu {
    border: red;
}
"""

    stylesheet = Stylesheet()
    stylesheet.add_source(bad_selector, None)

    with pytest.raises(TokenError) as error:
        stylesheet.parse()

    assert error.value.start == (2, 7)



"""Tests for the tooltips."""

from typing_extensions import Final

from textual.app import App, ComposeResult
from textual.widgets import Static


class TooltipApp(App[None]):
    TOOLTIP_DELAY = 0.4
    CSS = """
    Static {
        width: 1fr;
        height: 1fr;
    }
    """

    @staticmethod
    def tip(static: Static) -> Static:
        static.tooltip = "This is a test tooltip"
        return static

    def compose(self) -> ComposeResult:
        yield Static(id="mr-pink")
        yield self.tip(Static(id="mr-blue"))


TOOLTIP_TIMEOUT: Final[float] = 0.4 + 0.1
"""How long to wait for a tooltip to appear.

The 0.4 is the value defined with TOOLTIP_DELAY, and the 0.1 is a bit of
wiggle room.
"""


async def test_no_tip_gets_no_tooltip() -> None:
    """If there is no tooltip, none should show."""
    async with TooltipApp().run_test(tooltips=True) as pilot:
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.hover("#mr-pink")
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.pause(TOOLTIP_TIMEOUT)
        assert pilot.app.query_one("#textual-tooltip").display is False


async def test_tip_gets_a_tooltip() -> None:
    """If there is a tooltip, it should show."""
    async with TooltipApp().run_test(tooltips=True) as pilot:
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.hover("#mr-blue")
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.pause(TOOLTIP_TIMEOUT)
        assert pilot.app.query_one("#textual-tooltip").display is True


async def test_mouse_move_removes_a_tooltip() -> None:
    """If there is a mouse move when there is a tooltip, it should disappear."""
    async with TooltipApp().run_test(tooltips=True) as pilot:
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.hover("#mr-blue")
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.pause(TOOLTIP_TIMEOUT)
        assert pilot.app.query_one("#textual-tooltip").display is True
        await pilot.hover("#mr-pink")
        await pilot.pause(TOOLTIP_TIMEOUT)
        assert pilot.app.query_one("#textual-tooltip").display is False


async def test_removing_tipper_should_remove_tooltip() -> None:
    """If the tipping widget is removed, it should remove the tooltip."""
    async with TooltipApp().run_test(tooltips=True) as pilot:
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.hover("#mr-blue")
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.pause(TOOLTIP_TIMEOUT)
        assert pilot.app.query_one("#textual-tooltip").display is True
        await pilot.app.query_one("#mr-blue").remove()
        await pilot.pause()
        assert pilot.app.query_one("#textual-tooltip").display is False


async def test_making_tipper_invisible_should_remove_tooltip() -> None:
    """If the tipping widget is made invisible, it should remove the tooltip."""
    async with TooltipApp().run_test(tooltips=True) as pilot:
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.hover("#mr-blue")
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.pause(TOOLTIP_TIMEOUT)
        assert pilot.app.query_one("#textual-tooltip").display is True
        pilot.app.query_one("#mr-blue").visible = False
        await pilot.pause()
        assert pilot.app.query_one("#textual-tooltip").display is False


async def test_making_tipper_not_displayed_should_remove_tooltip() -> None:
    """If the tipping widget is made display none, it should remove the tooltip."""
    async with TooltipApp().run_test(tooltips=True) as pilot:
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.hover("#mr-blue")
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.pause(TOOLTIP_TIMEOUT)
        assert pilot.app.query_one("#textual-tooltip").display is True
        pilot.app.query_one("#mr-blue").display = False
        await pilot.pause()
        assert pilot.app.query_one("#textual-tooltip").display is False


async def test_making_tipper_shuffle_away_should_remove_tooltip() -> None:
    """If the tipping widget moves from under the cursor, it should remove the tooltip."""
    async with TooltipApp().run_test(tooltips=True) as pilot:
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.hover("#mr-blue")
        assert pilot.app.query_one("#textual-tooltip").display is False
        await pilot.pause(TOOLTIP_TIMEOUT)
        assert pilot.app.query_one("#textual-tooltip").display is True
        await pilot.app.mount(Static(id="mr-brown"), before="#mr-blue")
        await pilot.pause()
        assert pilot.app.query_one("#textual-tooltip").display is False



from __future__ import annotations

from textual.binding import Binding
from textual.reactive import reactive
from textual.widgets import Static


class Link(Static, can_focus=True):
    """A simple, clickable link that opens a URL."""

    DEFAULT_CSS = """
    Link {
        width: auto;
        height: auto;
        min-height: 1;
        color: $accent;
        text-style: underline;
        &:hover { color: $accent-lighten-1; }
        &:focus { text-style: bold reverse; }
    }
    """

    BINDINGS = [Binding("enter", "open_link", "Open link")]
    """
    | Key(s) | Description |
    | :- | :- |
    | enter | Open the link in the browser. |    
    """

    text: reactive[str] = reactive("", layout=True)
    url: reactive[str] = reactive("")

    def __init__(
        self,
        text: str,
        *,
        url: str | None = None,
        tooltip: str | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """A link widget.

        Args:
            text: Text of the link.
            url: A URL to open, when clicked. If `None`, the `text` parameter will also be used as the url.
            tooltip: Optional tooltip.
            name: Name of widget.
            id: ID of Widget.
            classes: Space separated list of class names.
            disabled: Whether the static is disabled or not.
        """
        super().__init__(
            text, name=name, id=id, classes=classes, disabled=disabled, markup=False
        )
        self.set_reactive(Link.text, text)
        self.set_reactive(Link.url, text if url is None else url)
        self.tooltip = tooltip

    def watch_text(self, text: str) -> None:
        self.update(text)

    def on_click(self) -> None:
        self.action_open_link()

    def action_open_link(self) -> None:
        if self.url:
            self.app.open_url(self.url)



from textual.app import App
from textual.containers import Container, Grid
from textual.widgets import Placeholder


class MarginAllApp(App):
    CSS_PATH = "margin_all.tcss"

    def compose(self):
        yield Grid(
            Container(Placeholder("no margin", id="p1"), classes="bordered"),
            Container(Placeholder("margin: 1", id="p2"), classes="bordered"),
            Container(Placeholder("margin: 1 5", id="p3"), classes="bordered"),
            Container(Placeholder("margin: 1 1 2 6", id="p4"), classes="bordered"),
            Container(Placeholder("margin-top: 4", id="p5"), classes="bordered"),
            Container(Placeholder("margin-right: 3", id="p6"), classes="bordered"),
            Container(Placeholder("margin-bottom: 4", id="p7"), classes="bordered"),
            Container(Placeholder("margin-left: 3", id="p8"), classes="bordered"),
        )


if __name__ == "__main__":
    app = MarginAllApp()
    app.run()



import csv
import io

from textual.app import App, ComposeResult
from textual.widgets import DataTable

CSV = """lane,swimmer,country,time
4,Joseph Schooling,Singapore,50.39
2,Michael Phelps,United States,51.14
5,Chad le Clos,South Africa,51.14
6,László Cseh,Hungary,51.14
3,Li Zhuhao,China,51.26
8,Mehdy Metella,France,51.58
7,Tom Shields,United States,51.73
1,Aleksandr Sadovnikov,Russia,51.84"""


class TableApp(App):
    def compose(self) -> ComposeResult:
        table = DataTable()
        table.focus()
        table.cursor_type = "row"
        table.fixed_columns = 1
        table.fixed_rows = 1
        yield table

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        rows = csv.reader(io.StringIO(CSV))
        table.add_columns(*next(rows))
        table.add_rows(rows)


if __name__ == "__main__":
    app = TableApp()
    app.run()



from __future__ import annotations

from time import sleep

from textual.notifications import Notification


def test_message() -> None:
    """A notification should not change the message."""
    assert Notification("test").message == "test"


def test_default_title() -> None:
    """A notification with no title should have a empty title."""
    assert Notification("test").title == ""


def test_default_severity_level() -> None:
    """The default severity level should be as expected."""
    assert Notification("test").severity == "information"


def test_default_timeout() -> None:
    """The default timeout should be as expected."""
    assert Notification("test").timeout == Notification.timeout


def test_identity_is_unique() -> None:
    """A collection of notifications should, by default, have unique IDs."""
    notifications: set[str] = set()
    for _ in range(1000):
        notifications.add(Notification("test").identity)
    assert len(notifications) == 1000


def test_time_out() -> None:
    test = Notification("test", timeout=0.5)
    assert test.has_expired is False
    sleep(0.6)
    assert test.has_expired is True



from __future__ import annotations

from textual.css.parse import substitute_references

__all__ = ["MarkupError", "escape", "to_content"]

import re
from string import Template
from typing import TYPE_CHECKING, Callable, Mapping, Match

from textual._context import active_app
from textual.color import Color
from textual.css.tokenize import (
    COLOR,
    PERCENT,
    TOKEN,
    VARIABLE_REF,
    Expect,
    TokenizerState,
    tokenize_values,
)
from textual.style import Style

if TYPE_CHECKING:
    from textual.content import Content


class MarkupError(Exception):
    """An error occurred parsing Textual markup."""


expect_markup_tag = (
    Expect(
        "markup style value",
        end_tag=r"(?<!\\)\]",
        key=r"[@a-zA-Z_-][a-zA-Z0-9_-]*=",
        percent=PERCENT,
        color=COLOR,
        token=TOKEN,
        variable_ref=VARIABLE_REF,
        whitespace=r"\s+",
    )
    .expect_eof()
    .expect_semicolon(False)
)

expect_markup = Expect(
    "markup tag",
    open_closing_tag=r"(?<!\\)\[/",
    open_tag=r"(?<!\\)\[",
).extract_text()

expect_markup_expression = (
    Expect(
        "markup value",
        end_tag=r"(?<!\\)\]",
        word=r"[\w\.]+",
        round_start=r"\(",
        round_end=r"\)",
        square_start=r"\[",
        square_end=r"\]",
        curly_start=r"\{",
        curly_end=r"\}",
        comma=",",
        whitespace=r"\s+",
        double_string=r"\".*?\"",
        single_string=r"'.*?'",
    )
    .expect_eof()
    .expect_semicolon(False)
)


class MarkupTokenizer(TokenizerState):
    """Tokenizes Textual markup."""

    EXPECT = expect_markup.expect_eof()
    STATE_MAP = {
        "open_tag": expect_markup_tag,
        "open_closing_tag": expect_markup_tag,
        "end_tag": expect_markup,
        "key": expect_markup_expression,
    }
    STATE_PUSH = {
        "round_start": expect_markup_expression,
        "square_start": expect_markup_expression,
        "curly_start": expect_markup_expression,
    }
    STATE_POP = {
        "round_end": "round_start",
        "square_end": "square_start",
        "curly_end": "curly_start",
    }


expect_style = Expect(
    "style token",
    end_tag=r"(?<!\\)\]",
    key=r"[@a-zA-Z_-][a-zA-Z0-9_-]*=",
    percent=PERCENT,
    color=COLOR,
    token=TOKEN,
    variable_ref=VARIABLE_REF,
    whitespace=r"\s+",
    double_string=r"\".*?\"",
    single_string=r"'.*?'",
).expect_semicolon(False)


class StyleTokenizer(TokenizerState):
    """Tokenizes a style"""

    EXPECT = expect_style.expect_eof()
    STATE_MAP = {"key": expect_markup_expression}
    STATE_PUSH = {
        "round_start": expect_markup_expression,
        "square_start": expect_markup_expression,
        "curly_start": expect_markup_expression,
    }


STYLES = {"bold", "dim", "italic", "underline", "reverse", "strike"}
STYLE_ABBREVIATIONS = {
    "b": "bold",
    "d": "dim",
    "i": "italic",
    "u": "underline",
    "r": "reverse",
    "s": "strike",
}

_ReStringMatch = Match[str]  # regex match object
_ReSubCallable = Callable[[_ReStringMatch], str]  # Callable invoked by re.sub
_EscapeSubMethod = Callable[[_ReSubCallable, str], str]  # Sub method of a compiled re


def escape(
    markup: str,
    _escape: _EscapeSubMethod = re.compile(r"(\\*)(\[[a-z#/@][^[]*?])").sub,
) -> str:
    """Escapes text so that it won't be interpreted as markup.

    Args:
        markup (str): Content to be inserted in to markup.

    Returns:
        str: Markup with square brackets escaped.
    """

    def escape_backslashes(match: Match[str]) -> str:
        """Called by re.sub replace matches."""
        backslashes, text = match.groups()
        return f"{backslashes}{backslashes}\\{text}"

    markup = _escape(escape_backslashes, markup)
    if markup.endswith("\\") and not markup.endswith("\\\\"):
        return markup + "\\"

    return markup


def parse_style(style: str, variables: dict[str, str] | None = None) -> Style:
    """Parse an encoded style.

    Args:
        style: Style encoded in a string.
        variables: Mapping of variables, or `None` to import from active app.

    Returns:
        A Style object.
    """

    styles: dict[str, bool | None] = {}
    color: Color | None = None
    background: Color | None = None
    is_background: bool = False
    style_state: bool = True

    tokenizer = StyleTokenizer()
    meta = {}

    if variables is None:
        try:
            app = active_app.get()
        except LookupError:
            reference_tokens = {}
        else:
            reference_tokens = app.stylesheet._variable_tokens
    else:
        reference_tokens = tokenize_values(variables)

    iter_tokens = iter(
        substitute_references(
            tokenizer(style, ("inline style", "")),
            reference_tokens,
        )
    )

    for token in iter_tokens:
        token_name = token.name
        token_value = token.value
        if token_name == "key":
            key = token_value.rstrip("=")
            parenthesis: list[str] = []
            value_text: list[str] = []
            first_token = next(iter_tokens)
            if first_token.name in {"double_string", "single_string"}:
                meta[key] = first_token.value[1:-1]
                break
            else:
                value_text.append(first_token.value)
                for token in iter_tokens:
                    if token.name == "whitespace" and not parenthesis:
                        break
                    value_text.append(token.value)
                    if token.name in {"round_start", "square_start", "curly_start"}:
                        parenthesis.append(token.value)
                    elif token.name in {"round_end", "square_end", "curly_end"}:
                        parenthesis.pop()
                        if not parenthesis:
                            break
                tokenizer.expect(StyleTokenizer.EXPECT)

                value = "".join(value_text)
                meta[key] = value

        elif token_name == "color":
            if is_background:
                background = Color.parse(token.value)
            else:
                color = Color.parse(token.value)

        elif token_name == "token":
            if token_value == "link":
                if "link" not in meta:
                    meta["link"] = ""
            elif token_value == "on":
                is_background = True
            elif token_value == "auto":
                if is_background:
                    background = Color.automatic()
                else:
                    color = Color.automatic()
            elif token_value == "not":
                style_state = False
            elif token_value in STYLES:
                styles[token_value] = style_state
                style_state = True
            elif token_value in STYLE_ABBREVIATIONS:
                styles[STYLE_ABBREVIATIONS[token_value]] = style_state
                style_state = True
            else:
                if is_background:
                    background = Color.parse(token_value)
                else:
                    color = Color.parse(token_value)

        elif token_name == "percent":
            percent = int(token_value.rstrip("%")) / 100.0
            if is_background:
                if background is not None:
                    background = background.multiply_alpha(percent)
            else:
                if color is not None:
                    color = color.multiply_alpha(percent)

    parsed_style = Style(background, color, link=meta.pop("link", None), **styles)

    if meta:
        parsed_style += Style.from_meta(meta)
    return parsed_style


def to_content(
    markup: str,
    style: str | Style = "",
    template_variables: Mapping[str, object] | None = None,
) -> Content:
    """Convert markup to Content.

    Args:
        markup: String containing markup.
        style: Optional base style.
        template_variables: Mapping of string.Template variables

    Raises:
        MarkupError: If the markup is invalid.

    Returns:
        Content that renders the markup.
    """
    _rich_traceback_omit = True
    try:
        return _to_content(markup, style, template_variables)
    except Exception as error:
        # Ensure all errors are wrapped in a MarkupError
        raise MarkupError(str(error)) from None


def _to_content(
    markup: str,
    style: str | Style = "",
    template_variables: Mapping[str, object] | None = None,
) -> Content:
    """Internal function to convert markup to Content.

    Args:
        markup: String containing markup.
        style: Optional base style.
        template_variables: Mapping of string.Template variables

    Raises:
        MarkupError: If the markup is invalid.

    Returns:
        Content that renders the markup.
    """

    from textual.content import Content, Span

    tokenizer = MarkupTokenizer()
    text: list[str] = []
    text_append = text.append
    iter_tokens = iter(tokenizer(markup, ("inline", "")))

    style_stack: list[tuple[int, str, str]] = []

    spans: list[Span] = []

    position = 0
    tag_text: list[str]

    normalize_markup_tag = Style._normalize_markup_tag

    if template_variables is None:
        process_text = lambda text: text

    else:

        def process_text(template_text: str, /) -> str:
            if "$" in template_text:
                return Template(template_text).safe_substitute(template_variables)
            return template_text

    for token in iter_tokens:
        token_name = token.name
        if token_name == "text":
            value = process_text(token.value.replace("\\[", "["))
            text_append(value)
            position += len(value)

        elif token_name == "open_tag":
            tag_text = []

            for token in iter_tokens:
                if token.name == "end_tag":
                    break
                tag_text.append(token.value)
            opening_tag = "".join(tag_text).strip()
            style_stack.append(
                (position, opening_tag, normalize_markup_tag(opening_tag))
            )

        elif token_name == "open_closing_tag":
            tag_text = []
            for token in iter_tokens:
                if token.name == "end_tag":
                    break
                tag_text.append(token.value)
            closing_tag = "".join(tag_text).strip()
            normalized_closing_tag = normalize_markup_tag(closing_tag)
            if normalized_closing_tag:
                for index, (tag_position, tag_body, normalized_tag_body) in enumerate(
                    reversed(style_stack), 1
                ):
                    if normalized_tag_body == normalized_closing_tag:
                        style_stack.pop(-index)
                        if tag_position != position:
                            spans.append(Span(tag_position, position, tag_body))
                        break
                else:
                    raise MarkupError(
                        f"closing tag '[/{closing_tag}]' does not match any open tag"
                    )

            else:
                if not style_stack:
                    raise MarkupError("auto closing tag ('[/]') has nothing to close")
                open_position, tag_body, _ = style_stack.pop()
                spans.append(Span(open_position, position, tag_body))

    content_text = "".join(text)
    text_length = len(content_text)
    while style_stack:
        position, tag_body, _ = style_stack.pop()
        spans.append(Span(position, text_length, tag_body))

    if style:
        content = Content(content_text, [Span(0, len(content_text), style), *spans])
    else:
        content = Content(content_text, spans)

    return content


if __name__ == "__main__":  # pragma: no cover
    from textual._markup_playground import MarkupPlayground

    app = MarkupPlayground()
    app.run()



from textual.app import App
from textual.widgets import Label


class VisibilityApp(App):
    CSS_PATH = "visibility.tcss"

    def compose(self):
        yield Label("Widget 1")
        yield Label("Widget 2", classes="invisible")
        yield Label("Widget 3")


if __name__ == "__main__":
    app = VisibilityApp()
    app.run()



from functools import partial

from textual.app import App
from textual.command import CommandPalette, Hit, Hits, Provider
from textual.widgets import Input


class SimpleSource(Provider):
    async def search(self, _: str) -> Hits:
        def goes_nowhere_does_nothing(selection: int) -> None:
            assert isinstance(self.app, CommandPaletteRunOnSelectApp)
            self.app.selection = selection

        for n in range(100):
            yield Hit(
                n + 1 / 100,
                str(n),
                partial(goes_nowhere_does_nothing, n),
                str(n),
                f"This is help for {n}",
            )


class CommandPaletteRunOnSelectApp(App[None]):
    COMMANDS = {SimpleSource}

    def __init__(self) -> None:
        super().__init__()
        self.selection: int | None = None


async def test_with_run_on_select_on() -> None:
    """With run on select on, the callable should be instantly run."""
    async with CommandPaletteRunOnSelectApp().run_test() as pilot:
        save = CommandPalette.run_on_select
        CommandPalette.run_on_select = True
        assert isinstance(pilot.app, CommandPaletteRunOnSelectApp)
        pilot.app.action_command_palette()
        await pilot.press("0")
        await pilot.app.screen.workers.wait_for_complete()
        await pilot.press("down")
        await pilot.press("enter")
        assert pilot.app.selection is not None
        CommandPalette.run_on_select = save


class CommandPaletteDoNotRunOnSelectApp(CommandPaletteRunOnSelectApp):
    def __init__(self) -> None:
        super().__init__()


async def test_with_run_on_select_off() -> None:
    """With run on select off, the callable should not be instantly run."""
    async with CommandPaletteDoNotRunOnSelectApp().run_test() as pilot:
        save = CommandPalette.run_on_select
        CommandPalette.run_on_select = False
        assert isinstance(pilot.app, CommandPaletteDoNotRunOnSelectApp)
        pilot.app.action_command_palette()
        await pilot.press("0")
        await pilot.app.screen.workers.wait_for_complete()
        await pilot.press("down")
        await pilot.press("enter")
        assert pilot.app.selection is None
        assert pilot.app.screen.query_one(Input).value != ""
        await pilot.press("enter")
        assert pilot.app.selection is not None
        CommandPalette.run_on_select = save



"""Tests the rendering of the TextArea for all supported languages."""
from textual.app import App, ComposeResult
from textual.widgets import TextArea


class TextAreaSnapshot(App):
    def compose(self) -> ComposeResult:
        text_area = TextArea.code_editor()
        text_area.cursor_blink = False
        yield text_area


app = TextAreaSnapshot()
if __name__ == "__main__":
    app.run()



from textual.app import App, ComposeResult
from textual.color import Color
from textual.containers import Container
from textual.css.scalar import ScalarOffset


class StyleApp(App[None]):
    CSS = """
    Container {
        border: round green !important;
        outline: round green !important;
        align: right bottom !important;
        content-align: right bottom !important;
        offset: 17 23 !important;
        overflow: hidden hidden !important;
        padding: 10 20 30 40 !important;
        scrollbar-size: 23 42 !important;
    }

    Container.more-specific {
        border: solid red;
        outline: solid red;
        align: center middle;
        content-align: center middle;
        offset: 0 0;
        overflow: scroll scroll;
        padding: 1 2 3 4;
        scrollbar-size: 1 2;
    }
    """

    def compose(self) -> ComposeResult:
        yield Container(classes="more-specific")


async def test_border_importance():
    """Border without sides should support !important"""
    async with StyleApp().run_test() as pilot:
        border = pilot.app.query_one(Container).styles.border
        desired = ("round", Color.parse("green"))
        assert border.top == desired
        assert border.left == desired
        assert border.bottom == desired
        assert border.right == desired


async def test_outline_importance():
    """Outline without sides should support !important"""
    async with StyleApp().run_test() as pilot:
        outline = pilot.app.query_one(Container).styles.outline
        desired = ("round", Color.parse("green"))
        assert outline.top == desired
        assert outline.left == desired
        assert outline.bottom == desired
        assert outline.right == desired


async def test_align_importance():
    """Align without direction should support !important"""
    async with StyleApp().run_test() as pilot:
        assert pilot.app.query_one(Container).styles.align == ("right", "bottom")


async def test_content_align_importance():
    """Content align without direction should support !important"""
    async with StyleApp().run_test() as pilot:
        assert pilot.app.query_one(Container).styles.content_align == (
            "right",
            "bottom",
        )


async def test_offset_importance():
    """Offset without direction should support !important"""
    async with StyleApp().run_test() as pilot:
        assert pilot.app.query_one(Container).styles.offset == ScalarOffset.from_offset(
            (17, 23)
        )


async def test_overflow_importance():
    """Overflow without direction should support !important"""
    async with StyleApp().run_test() as pilot:
        assert pilot.app.query_one(Container).styles.overflow_x == "hidden"
        assert pilot.app.query_one(Container).styles.overflow_y == "hidden"


async def test_padding_importance():
    """Padding without sides should support !important"""
    async with StyleApp().run_test() as pilot:
        padding = pilot.app.query_one(Container).styles.padding
        assert padding.top == 10
        assert padding.left == 40
        assert padding.bottom == 30
        assert padding.right == 20


async def test_scrollbar_size_importance():
    """Scrollbar size without direction should support !important"""
    async with StyleApp().run_test() as pilot:
        assert pilot.app.query_one(Container).styles.scrollbar_size_horizontal == 23
        assert pilot.app.query_one(Container).styles.scrollbar_size_vertical == 42



from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.widgets import Label


class BackgroundTintApp(App):
    CSS_PATH = "background_tint.tcss"

    def compose(self) -> ComposeResult:
        with Vertical(id="tint1"):
            yield Label("0%")
        with Vertical(id="tint2"):
            yield Label("25%")
        with Vertical(id="tint3"):
            yield Label("50%")
        with Vertical(id="tint4"):
            yield Label("75%")
        with Vertical(id="tint5"):
            yield Label("100%")


if __name__ == "__main__":
    app = BackgroundTintApp()
    app.run()



from __future__ import annotations

from typing import TYPE_CHECKING

from rich.console import RenderableType
from rich.protocol import is_renderable

if TYPE_CHECKING:
    from textual.app import RenderResult

from textual.errors import RenderError
from textual.visual import SupportsVisual, Visual, VisualType, visualize
from textual.widget import Widget


def _check_renderable(renderable: object):
    """Check if a renderable conforms to the Rich Console protocol
    (https://rich.readthedocs.io/en/latest/protocol.html)

    Args:
        renderable: A potentially renderable object.

    Raises:
        RenderError: If the object can not be rendered.
    """
    if not is_renderable(renderable) and not hasattr(renderable, "visualize"):
        raise RenderError(
            f"unable to render {renderable.__class__.__name__!r} type; must be a str, Text, Rich renderable oor Textual Visual instance"
        )


class Static(Widget, inherit_bindings=False):
    """A widget to display simple static content, or use as a base class for more complex widgets.

    Args:
        content: A Rich renderable, or string containing console markup.
        expand: Expand content if required to fill container.
        shrink: Shrink content if required to fill container.
        markup: True if markup should be parsed and rendered.
        name: Name of widget.
        id: ID of Widget.
        classes: Space separated list of class names.
        disabled: Whether the static is disabled or not.
    """

    DEFAULT_CSS = """
    Static {
        height: auto;
    }
    """

    _renderable: RenderableType | SupportsVisual

    def __init__(
        self,
        content: RenderableType | SupportsVisual = "",
        *,
        expand: bool = False,
        shrink: bool = False,
        markup: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            name=name, id=id, classes=classes, disabled=disabled, markup=markup
        )
        self.expand = expand
        self.shrink = shrink
        self._content = content
        self._visual: Visual | None = None

    @property
    def visual(self) -> Visual:
        if self._visual is None:
            self._visual = visualize(self, self._content, markup=self._render_markup)
        return self._visual

    @property
    def renderable(self) -> RenderableType | SupportsVisual:
        return self._content or ""

    @renderable.setter
    def renderable(self, renderable: RenderableType | SupportsVisual) -> None:
        self._renderable = renderable
        self._visual = None
        self.clear_cached_dimensions()

    def render(self) -> RenderResult:
        """Get a rich renderable for the widget's content.

        Returns:
            A rich renderable.
        """
        return self.visual

    def update(self, content: VisualType = "") -> None:
        """Update the widget's content area with new text or Rich renderable.

        Args:
            content: New content.
        """

        self._content = content
        self._visual = visualize(self, content, markup=self._render_markup)
        self.refresh(layout=True)



from textual.app import App, ComposeResult
from textual.widgets import Link


class LabelApp(App):
    AUTO_FOCUS = None
    CSS = """
    Screen {
        align: center middle;
    }
    """

    def compose(self) -> ComposeResult:
        yield Link(
            "Go to textualize.io",
            url="https://textualize.io",
            tooltip="Click me",
        )


if __name__ == "__main__":
    app = LabelApp()
    app.run()



from rich.segment import Segment

from textual.app import App, ComposeResult
from textual.widgets import Header, Footer
from textual.scroll_view import ScrollView
from textual.strip import Strip
from textual.geometry import Size


class TestScrollView(ScrollView, can_focus=True):
    def __init__(self, height: int, border_title: str) -> None:
        super().__init__()
        self.virtual_size = Size(0, height)
        self.border_title = border_title

    def render_line(self, y: int) -> Strip:
        return Strip(
            [
                Segment(f"Welcome to line {self.scroll_offset.y + y}"),
            ]
        )


class ScrollViewTester(App[None]):
    """Check the scrollbar fits the end."""

    CSS = """
    TestScrollView {
        background: $primary-darken-2;
        border: round red;
    }
    """

    def compose(self) -> ComposeResult:
        yield Header()
        yield TestScrollView(height=1000, border_title=f"1")
        yield Footer()

    def on_ready(self) -> None:
        self.query_one(TestScrollView).scroll_end(animate=False)


if __name__ == "__main__":
    ScrollViewTester().run()



from __future__ import annotations

import io
from collections import deque
from typing import (
    Callable,
    Deque,
    Generator,
    Generic,
    Iterable,
    NamedTuple,
    Tuple,
    TypeVar,
)

from typing_extensions import TypeAlias


class ParseError(Exception):
    """Parse related errors."""


class ParseEOF(ParseError):
    """End of Stream."""


class Awaitable:
    """Base class for an parser awaitable."""

    __slots__: list[str] = []


class _Read(Awaitable):
    """Read a predefined number of bytes."""

    __slots__ = ["remaining"]

    def __init__(self, count: int) -> None:
        self.remaining = count


class _Read1(Awaitable):
    """Read a single byte."""

    __slots__: list[str] = []


TokenType = TypeVar("TokenType")

ByteStreamTokenCallback: TypeAlias = Callable[[TokenType], None]


class ByteStreamParser(Generic[TokenType]):
    """A parser to feed in binary data and generate a sequence of tokens."""

    read = _Read
    read1 = _Read1

    def __init__(self) -> None:
        """Initialize the parser."""
        self._buffer = io.BytesIO()
        self._eof = False
        self._tokens: Deque[TokenType] = deque()
        self._gen = self.parse(self._tokens.append)
        self._awaiting: Awaitable | TokenType = next(self._gen)

    @property
    def is_eof(self) -> bool:
        """Is the parser at the end of file?"""
        return self._eof

    def feed(self, data: bytes) -> Iterable[TokenType]:
        """Feed the parser some data, return an iterable of tokens."""
        if self._eof:
            raise ParseError("end of file reached") from None
        if not data:
            self._eof = True
            try:
                self._gen.send(self._buffer.getvalue())
            except StopIteration:
                raise ParseError("end of file reached") from None
            while self._tokens:
                yield self._tokens.popleft()

            self._buffer.truncate(0)
            return

        _buffer = self._buffer
        pos = 0
        tokens = self._tokens
        popleft = tokens.popleft
        data_size = len(data)

        while tokens:
            yield popleft()

        while pos < data_size:
            _awaiting = self._awaiting
            if isinstance(_awaiting, _Read1):
                self._awaiting = self._gen.send(data[pos : pos + 1])
                pos += 1
            elif isinstance(_awaiting, _Read):
                remaining = _awaiting.remaining
                chunk = data[pos : pos + remaining]
                chunk_size = len(chunk)
                pos += chunk_size
                _buffer.write(chunk)
                remaining -= chunk_size
                if remaining:
                    _awaiting.remaining = remaining
                else:
                    self._awaiting = self._gen.send(_buffer.getvalue())
                    _buffer.seek(0)
                    _buffer.truncate()

            while tokens:
                yield popleft()

    def parse(
        self, on_token: ByteStreamTokenCallback
    ) -> Generator[Awaitable, bytes, None]:
        """Implement in a sub-class to define parse behavior.

        Args:
            on_token: A callable which accepts the token type, and returns None.

        """
        yield from ()


class BytePacket(NamedTuple):
    """A type and payload."""

    type: str
    payload: bytes


class ByteStream(ByteStreamParser[Tuple[str, bytes]]):
    """A stream of packets in the following format.

    1 Byte for the type.
    4 Bytes for the big endian encoded size
    Arbitrary payload

    """

    def parse(
        self, on_token: ByteStreamTokenCallback
    ) -> Generator[Awaitable, bytes, None]:
        read1 = self.read1
        read = self.read
        from_bytes = int.from_bytes
        while not self.is_eof:
            packet_type = (yield read1()).decode("utf-8", "ignore")
            size = from_bytes((yield read(4)), "big")
            payload = (yield read(size)) if size else b""
            on_token(BytePacket(packet_type, payload))



"""

The compositor handles combining widgets into a single screen (i.e. compositing).

It also stores the results of that process, so that Textual knows the widgets on
the screen and their locations. The compositor uses this information to answer
queries regarding the widget under an offset, or the style under an offset.

Additionally, the compositor can render portions of the screen which may have updated,
without having to render the entire screen.
"""

from __future__ import annotations

from operator import itemgetter
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterable,
    Mapping,
    NamedTuple,
    Sequence,
    cast,
)

import rich.repr
from rich.console import Console, ConsoleOptions, RenderableType, RenderResult
from rich.control import Control
from rich.segment import Segment
from rich.style import Style

from textual import errors
from textual._cells import cell_len
from textual._context import visible_screen_stack
from textual._loop import loop_last
from textual.geometry import NULL_SPACING, Offset, Region, Size, Spacing
from textual.map_geometry import MapGeometry
from textual.strip import Strip, StripRenderable

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from textual.screen import Screen
    from textual.widget import Widget


class ReflowResult(NamedTuple):
    """The result of a reflow operation. Describes the chances to widgets."""

    hidden: set[Widget]  # Widgets that are hidden
    shown: set[Widget]  # Widgets that are shown
    resized: set[Widget]  # Widgets that have been resized


# Maps a widget on to its geometry (information that describes its position in the composition)
CompositorMap: TypeAlias = "dict[Widget, MapGeometry]"


class CompositorUpdate:
    """An update generated by the compositor, which also doubles as console renderables."""

    def render_segments(self, console: Console) -> str:
        """Render the update to raw data, suitable for writing to terminal.

        Args:
            console: Console instance.

        Returns:
            Raw data with escape sequences.
        """
        return ""


@rich.repr.auto(angular=True)
class LayoutUpdate(CompositorUpdate):
    """A renderable containing the result of a render for a given region."""

    def __init__(self, strips: list[Strip], region: Region) -> None:
        self.strips = strips
        self.region = region

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        x = self.region.x
        new_line = Segment.line()
        move_to = Control.move_to
        for last, (y, line) in loop_last(enumerate(self.strips, self.region.y)):
            yield move_to(x, y).segment
            yield from line
            if not last:
                yield new_line

    def render_segments(self, console: Console) -> str:
        """Render the update to raw data, suitable for writing to terminal.

        Args:
            console: Console instance.

        Returns:
            Raw data with escape sequences.
        """
        sequences: list[str] = []
        append = sequences.append
        x = self.region.x
        move_to = Control.move_to
        for last, (y, line) in loop_last(enumerate(self.strips, self.region.y)):
            append(move_to(x, y).segment.text)
            append(line.render(console))
            if not last:
                append("\n")
        return "".join(sequences)

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.region


@rich.repr.auto(angular=True)
class InlineUpdate(CompositorUpdate):
    """A renderable to write an inline update."""

    def __init__(self, strips: list[Strip], clear: bool = False) -> None:
        self.strips = strips
        self.clear = clear

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        new_line = Segment.line()
        for last, line in loop_last(self.strips):
            yield from line
            if not last:
                yield new_line

    def render_segments(self, console: Console) -> str:
        """Render the update to raw data, suitable for writing to terminal.

        Args:
            console: Console instance.

        Returns:
            Raw data with escape sequences.
        """
        sequences: list[str] = []
        append = sequences.append
        for last, strip in loop_last(self.strips):
            append(strip.render(console))
            if not last:
                append("\n")
        if self.clear:
            if len(self.strips) > 1:
                append("\n")
            append("\x1b[J")  # Clear down
        if len(self.strips) > 1:
            back_lines = len(self.strips) if self.clear else len(self.strips) - 1
            append(f"\x1b[{back_lines}A\r")  # Move cursor back to original position
        else:
            append("\r")
        append("\x1b[6n")  # Query new cursor position
        return "".join(sequences)


@rich.repr.auto(angular=True)
class ChopsUpdate(CompositorUpdate):
    """A renderable that applies updated spans to the screen."""

    def __init__(
        self,
        chops: Sequence[Mapping[int, Strip | None]],
        spans: list[tuple[int, int, int]],
        chop_ends: list[list[int]],
    ) -> None:
        """A renderable which updates chops (fragments of lines).

        Args:
            chops: A mapping of offsets to list of segments, per line.
            crop: Region to restrict update to.
            chop_ends: A list of the end offsets for each line
        """
        self.chops = chops
        self.spans = spans
        self.chop_ends = chop_ends

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        move_to = Control.move_to
        new_line = Segment.line()
        chops = self.chops
        chop_ends = self.chop_ends
        last_y = self.spans[-1][0]

        _cell_len = cell_len
        for y, x1, x2 in self.spans:
            line = chops[y]
            ends = chop_ends[y]
            for end, (x, strip) in zip(ends, line.items()):
                # TODO: crop to x extents
                if strip is None:
                    continue

                if x > x2 or end <= x1:
                    continue

                if x2 > x >= x1 and end <= x2:
                    yield move_to(x, y).segment
                    yield from strip
                    continue

                iter_segments = iter(strip)
                if x < x1:
                    for segment in iter_segments:
                        next_x = x + _cell_len(segment.text)
                        if next_x > x1:
                            yield move_to(x, y).segment
                            yield segment
                            break
                        x = next_x
                else:
                    yield move_to(x, y).segment
                if end <= x2:
                    yield from iter_segments
                else:
                    for segment in iter_segments:
                        if x >= x2:
                            break
                        yield segment
                        x += _cell_len(segment.text)

            if y != last_y:
                yield new_line

    def render_segments(self, console: Console) -> str:
        """Render the update to raw data, suitable for writing to terminal.

        Args:
            console: Console instance.

        Returns:
            Raw data with escape sequences.
        """

        sequences: list[str] = []
        append = sequences.append

        move_to = Control.move_to
        chops = self.chops
        chop_ends = self.chop_ends
        last_y = self.spans[-1][0]

        for y, x1, x2 in self.spans:
            line = chops[y]
            ends = chop_ends[y]
            for end, (x, strip) in zip(ends, line.items()):
                if strip is None:
                    continue

                if x > x2 or end <= x1:
                    continue

                if x2 > x >= x1 and end <= x2:
                    append(move_to(x, y).segment.text)
                    append(strip.render(console))
                    continue

                strip = strip.crop(0, min(end, x2) - x)
                append(move_to(x, y).segment.text)
                append(strip.render(console))

            if y != last_y:
                append("\n")

        terminal_sequences = "".join(sequences)
        return terminal_sequences

    def __rich_repr__(self) -> rich.repr.Result:
        yield from ()


@rich.repr.auto(angular=True)
class Compositor:
    """Responsible for storing information regarding the relative positions of Widgets and rendering them."""

    def __init__(self) -> None:
        # A mapping of Widget on to its "render location" (absolute position / depth)
        self._full_map: CompositorMap = {}
        self._full_map_invalidated = True
        self._visible_map: CompositorMap | None = None
        self._layers: list[tuple[Widget, MapGeometry]] | None = None

        # All widgets considered in the arrangement
        # Note this may be a superset of self.full_map.keys() as some widgets may be invisible for various reasons
        self.widgets: set[Widget] = set()

        # Mapping of visible widgets on to their region, and clip region
        self._visible_widgets: dict[Widget, tuple[Region, Region]] | None = None

        # The top level widget
        self.root: Widget | None = None

        # Dimensions of the arrangement
        self.size = Size(0, 0)

        # The points in each line where the line bisects the left and right edges of the widget
        self._cuts: list[list[int]] | None = None

        # Regions that require an update
        self._dirty_regions: set[Region] = set()

        # Mapping of line numbers on to lists of widget and regions
        self._layers_visible: list[list[tuple[Widget, Region, Region]]] | None = None

    def clear(self) -> None:
        """Remove all references to widgets (used when the screen closes)."""
        self._full_map.clear()
        self._visible_map = None
        self._layers = None
        self.widgets.clear()
        self._visible_widgets = None
        self._layers_visible = None

    @classmethod
    def _regions_to_spans(
        cls, regions: Iterable[Region]
    ) -> Iterable[tuple[int, int, int]]:
        """Converts the regions to horizontal spans. Spans will be combined if they overlap
        or are contiguous to produce optimal non-overlapping spans.

        Args:
            regions: An iterable of Regions.

        Returns:
            Yields tuples of (Y, X1, X2).
        """
        inline_ranges: dict[int, list[tuple[int, int]]] = {}
        setdefault = inline_ranges.setdefault
        for region_x, region_y, width, height in regions:
            span = (region_x, region_x + width)
            for y in range(region_y, region_y + height):
                setdefault(y, []).append(span)

        slice_remaining = slice(1, None)
        for y, ranges in sorted(inline_ranges.items()):
            if len(ranges) == 1:
                # Special case of 1 span
                yield (y, *ranges[0])
            else:
                ranges.sort()
                x1, x2 = ranges[0]
                for next_x1, next_x2 in ranges[slice_remaining]:
                    if next_x1 <= x2:
                        if next_x2 > x2:
                            x2 = next_x2
                    else:
                        yield (y, x1, x2)
                        x1 = next_x1
                        x2 = next_x2
                yield (y, x1, x2)

    def __rich_repr__(self) -> rich.repr.Result:
        yield "size", self.size
        yield "widgets", self.widgets

    def reflow(self, parent: Widget, size: Size) -> ReflowResult:
        """Reflow (layout) widget and its children.

        Args:
            parent: The root widget.
            size: Size of the area to be filled.

        Returns:
            Hidden, shown, and resized widgets.
        """
        self._cuts = None
        self._layers = None
        self._layers_visible = None
        self._visible_widgets = None
        self._visible_map = None
        self.root = parent
        self.size = size

        # Keep a copy of the old map because we're going to compare it with the update
        old_map = self._full_map
        old_widgets = old_map.keys()

        map, widgets = self._arrange_root(parent, size, visible_only=False)

        new_widgets = map.keys()

        # Newly visible widgets
        shown_widgets = new_widgets - old_widgets

        # Newly hidden widgets
        hidden_widgets = self.widgets - widgets

        # Replace map and widgets
        self._full_map = map
        self.widgets = widgets

        # Contains widgets + geometry for every widget that changed (added, removed, or updated)
        changes = map.items() ^ old_map.items()

        # Widgets in both new and old
        common_widgets = old_widgets & new_widgets

        # Mark dirty regions.
        screen_region = size.region
        if screen_region not in self._dirty_regions:
            regions = {
                region
                for region in (
                    map_geometry.clip.intersection(map_geometry.region)
                    for _, map_geometry in changes
                )
                if region
            }
            self._dirty_regions.update(regions)

        resized_widgets = {
            widget
            for widget, (region, *_) in changes
            if (widget in common_widgets and old_map[widget].region[2:] != region[2:])
        }
        return ReflowResult(
            hidden=hidden_widgets,
            shown=shown_widgets,
            resized=resized_widgets,
        )

    def reflow_visible(self, parent: Widget, size: Size) -> set[Widget]:
        """Reflow only the visible children.

        This is a fast-path for scrolling.

        Args:
            parent: The root widget.
            size: Size of the area to be filled.

        Returns:
            Set of widgets that were exposed by the scroll.
        """
        self._cuts = None
        self._layers = None
        self._layers_visible = None
        self._visible_widgets = None
        self._full_map_invalidated = True
        self.root = parent
        self.size = size

        # Keep a copy of the old map because we're going to compare it with the update
        old_map = self._visible_map or {}
        map, widgets = self._arrange_root(parent, size, visible_only=True)

        # Replace map and widgets
        self._visible_map = map
        self.widgets = widgets

        exposed_widgets = map.keys() - old_map.keys()

        # Contains widgets + geometry for every widget that changed (added, removed, or updated)
        changes = map.items() ^ old_map.items()

        # Mark dirty regions.
        screen_region = size.region
        if screen_region not in self._dirty_regions:
            regions = {
                region
                for region in (
                    map_geometry.clip.intersection(map_geometry.region)
                    for _, map_geometry in changes
                )
                if region
            }
            self._dirty_regions.update(regions)

        return exposed_widgets

    @property
    def full_map(self) -> CompositorMap:
        """Lazily built compositor map that covers all widgets."""

        if self.root is None:
            return {}
        if self._full_map_invalidated:
            self._full_map_invalidated = False
            map, _widgets = self._arrange_root(self.root, self.size, visible_only=False)
            # Update any widgets which became visible in the interim
            self._full_map = map
            self._visible_widgets = None
            self._visible_map = None

        return self._full_map

    @property
    def visible_widgets(self) -> dict[Widget, tuple[Region, Region]]:
        """Get a mapping of widgets on to region and clip.

        Returns:
            Visible widget mapping.
        """

        if self._visible_widgets is None:
            map = (
                self._visible_map
                if self._visible_map is not None
                else (self._full_map or {})
            )
            screen = self.size.region
            in_screen = screen.overlaps
            overlaps = Region.overlaps

            # Widgets and regions in render order
            visible_widgets = [
                (order, widget, region, clip)
                for widget, (region, order, clip, _, _, _, _) in map.items()
                if in_screen(region) and overlaps(clip, region)
            ]
            visible_widgets.sort(key=itemgetter(0), reverse=True)
            self._visible_widgets = {
                widget: (region, clip) for _, widget, region, clip in visible_widgets
            }
        return self._visible_widgets

    def _arrange_root(
        self, root: Widget, size: Size, visible_only: bool = True
    ) -> tuple[CompositorMap, set[Widget]]:
        """Arrange a widget's children based on its layout attribute.

        Args:
            root: Top level widget.
            size: Size of visible area (screen).
            visible_only: Only update visible widgets (used in scrolling).

        Returns:
            Compositor map and set of widgets.
        """

        map: CompositorMap = {}
        widgets: set[Widget] = set()
        add_new_widget = widgets.add
        invisible_widgets: set[Widget] = set()
        add_new_invisible_widget = invisible_widgets.add
        layer_order: int = 0

        no_clip = size.region

        def add_widget(
            widget: Widget,
            virtual_region: Region,
            region: Region,
            order: tuple[tuple[int, int, int], ...],
            layer_order: int,
            clip: Region,
            visible: bool,
            dock_gutter: Spacing,
            _MapGeometry: type[MapGeometry] = MapGeometry,
        ) -> None:
            """Called recursively to place a widget and its children in the map.

            Args:
                widget: The widget to add.
                virtual_region: The Widget region relative to its container.
                region: The region the widget will occupy.
                order: Painting order information.
                layer_order: The order of the widget in its layer.
                clip: The clipping region (i.e. the viewport which contains it).
                visible: Whether the widget should be visible by default.
                    This may be overridden by the CSS rule `visibility`.
            """
            if not widget._is_mounted:
                return
            styles = widget.styles

            if (visibility := styles.get_rule("visibility")) is not None:
                visible = visibility == "visible"

            if visible:
                add_new_widget(widget)
            else:
                add_new_invisible_widget(widget)

            # Container region is minus border
            container_region = region.shrink(styles.gutter)
            container_size = container_region.size

            # Widgets with scrollbars (containers or scroll view) require additional processing
            if widget.is_scrollable:
                # The region that contains the content (container region minus scrollbars)
                child_region = (
                    container_region
                    if widget.loading
                    else widget._get_scrollable_region(container_region)
                )

                # The region covered by children relative to parent widget
                total_region = child_region.reset_offset

                if widget.is_container:
                    # Arrange the layout
                    arrange_result = widget._arrange(child_region.size)

                    arranged_widgets = arrange_result.widgets
                    widgets.update(arranged_widgets)

                    # Get the region that will be updated
                    sub_clip = clip.intersection(child_region)

                    if visible_only:
                        placements = arrange_result.get_visible_placements(
                            sub_clip - child_region.offset + widget.scroll_offset
                        )
                    else:
                        placements = arrange_result.placements
                    total_region = total_region.union(arrange_result.total_region)

                    # An offset added to all placements
                    placement_offset = container_region.offset
                    placement_scroll_offset = placement_offset - widget.scroll_offset

                    placements = [
                        placement.process_offset(size.region, placement_scroll_offset)
                        for placement in placements
                    ]

                    layers_to_index = {
                        layer_name: index
                        for index, layer_name in enumerate(widget.layers)
                    }

                    get_layer_index = layers_to_index.get

                    if widget._cover_widget is not None:
                        map[widget._cover_widget] = _MapGeometry(
                            region.shrink(widget.styles.gutter),
                            order,
                            clip,
                            region.size,
                            container_size,
                            virtual_region,
                            dock_gutter,
                        )

                    # Add all the widgets
                    for (
                        sub_region,
                        sub_region_offset,
                        _,
                        sub_widget,
                        z,
                        fixed,
                        overlay,
                        absolute,
                    ) in reversed(placements):
                        layer_index = get_layer_index(sub_widget.layer, 0)
                        # Combine regions with children to calculate the "virtual size"
                        if fixed:
                            widget_region = (
                                sub_region + sub_region_offset + placement_offset
                            )
                        else:
                            widget_region = (
                                sub_region + sub_region_offset + placement_scroll_offset
                            )

                        widget_order = order + ((layer_index, z, layer_order),)

                        if widget._cover_widget is None:
                            add_widget(
                                sub_widget,
                                sub_region,
                                widget_region,
                                ((1, 0, 0),) if overlay else widget_order,
                                layer_order,
                                no_clip if overlay else sub_clip,
                                visible,
                                arrange_result.scroll_spacing,
                            )
                        layer_order -= 1

                if visible:
                    # Add any scrollbars
                    if (
                        widget.show_vertical_scrollbar
                        or widget.show_horizontal_scrollbar
                    ):
                        for chrome_widget, chrome_region in widget._arrange_scrollbars(
                            container_region
                        ):
                            map[chrome_widget] = _MapGeometry(
                                chrome_region,
                                order,
                                clip,
                                container_size,
                                container_size,
                                chrome_region,
                                dock_gutter,
                            )

                    map[widget] = _MapGeometry(
                        region,
                        order,
                        clip,
                        total_region.size,
                        container_size,
                        virtual_region,
                        dock_gutter,
                    )

            elif visible:
                # Add the widget to the map
                map[widget._render_widget] = _MapGeometry(
                    region,
                    order,
                    clip,
                    region.size,
                    container_size,
                    virtual_region,
                    dock_gutter,
                )

        # Add top level (root) widget
        add_widget(
            root,
            size.region,
            size.region,
            ((0, 0, 0),),
            layer_order,
            size.region,
            True,
            NULL_SPACING,
        )
        widgets -= invisible_widgets
        return map, widgets

    @property
    def layers(self) -> list[tuple[Widget, MapGeometry]]:
        """Get widgets and geometry in layer order."""
        map = self._visible_map if self._visible_map is not None else self._full_map
        if self._layers is None:
            self._layers = sorted(
                map.items(), key=lambda item: item[1].order, reverse=True
            )
        return self._layers

    @property
    def layers_visible(self) -> list[list[tuple[Widget, Region, Region]]]:
        """Visible widgets and regions in layers order.

        Returns:
            Lists visible widgets per layer. Widgets are give as a tuple of
            (WIDGET, CROPPED_REGION, REGION). CROPPED_REGION is clipped by
            the container.

        """

        if self._layers_visible is None:
            layers_visible: list[list[tuple[Widget, Region, Region]]]
            layers_visible = [[] for y in range(self.size.height)]
            layers_visible_appends = [layer.append for layer in layers_visible]
            intersection = Region.intersection
            _range = range
            for widget, (region, clip) in self.visible_widgets.items():
                cropped_region = intersection(region, clip)
                _x, region_y, _width, region_height = cropped_region
                if region_height:
                    widget_location = (widget, cropped_region, region)
                    for y in _range(region_y, region_y + region_height):
                        layers_visible_appends[y](widget_location)
            self._layers_visible = layers_visible
        return self._layers_visible

    def __contains__(self, widget: Widget) -> bool:
        """Check if the widget was included in the last update.

        Args:
            widget: A widget.

        Returns:
            `True` if the widget was in the last refresh, or `False` if it wasn't.
        """
        # Try to avoid a recalculation of full_map if possible.
        return (
            widget in self.widgets
            or (self._visible_map is not None and widget in self._visible_map)
            or widget in self.full_map
        )

    def get_offset(self, widget: Widget) -> Offset:
        """Get the offset of a widget.

        Args:
            widget: Widget to query.

        Returns:
            Offset of widget.
        """
        try:
            if self._visible_map is not None:
                try:
                    return self._visible_map[widget].region.offset
                except KeyError:
                    pass
            return self.full_map[widget].region.offset
        except KeyError:
            raise errors.NoWidget("Widget is not in layout")

    def get_widget_at(self, x: int, y: int) -> tuple[Widget, Region]:
        """Get the widget under a given coordinate.

        Args:
            x: X Coordinate.
            y: Y Coordinate.

        Raises:
            errors.NoWidget: If there is not widget underneath (x, y).

        Returns:
            A tuple of the widget and its region.
        """

        contains = Region.contains
        if len(self.layers_visible) > y >= 0:
            for widget, cropped_region, region in self.layers_visible[int(y)]:
                if contains(cropped_region, x, y) and widget.visible:
                    return widget, region
        raise errors.NoWidget(f"No widget under screen coordinate ({x}, {y})")

    def get_widgets_at(self, x: int, y: int) -> Iterable[tuple[Widget, Region]]:
        """Get all widgets under a given coordinate.

        Args:
            x: X coordinate.
            y: Y coordinate.

        Returns:
            Sequence of (WIDGET, REGION) tuples.
        """
        contains = Region.contains
        for widget, cropped_region, region in self.layers_visible[y]:
            if contains(cropped_region, x, y) and widget.visible:
                yield widget, region

    def get_style_at(self, x: int, y: int) -> Style:
        """Get the Style at the given cell or Style.null()

        Args:
            x: X position within the Layout.
            y: Y position within the Layout.

        Returns:
            The Style at the cell (x, y) within the Layout.
        """
        try:
            widget, region = self.get_widget_at(x, y)
        except errors.NoWidget:
            return Style.null()
        if widget not in self.visible_widgets:
            return Style.null()

        x -= region.x
        y -= region.y

        visible_screen_stack.set(widget.app._background_screens)
        lines = widget.render_lines(Region(0, y, region.width, 1))

        if not lines:
            return Style.null()
        end = 0

        for segment in lines[0]:
            end += segment.cell_length
            if x < end:
                return segment.style or Style.null()

        return Style.null()

    def get_widget_and_offset_at(
        self, x: int, y: int
    ) -> tuple[Widget | None, Offset | None]:
        """Get the Style at the given cell, the offset within the content.

        Args:
            x: X position within the Layout.
            y: Y position within the Layout.

        Returns:
            A tuple of the widget at (x, y) and the offset within the widget.
        """
        try:
            widget, region = self.get_widget_at(x, y)
        except errors.NoWidget:
            return None, None
        if widget not in self.visible_widgets:
            return None, None

        if y >= widget.content_region.bottom:
            x, y = widget.content_region.bottom_right_inclusive

        gutter_left, gutter_right = widget.gutter.top_left
        x -= region.x + gutter_left
        y -= region.y + gutter_right

        visible_screen_stack.set(widget.app._background_screens)
        line = widget.render_line(y)

        end = 0
        start = 0
        offset_y: int | None = None
        offset_x = 0
        offset_x2 = 0

        for segment in line:
            end += len(segment.text)
            style = segment.style
            if style is not None and style._meta is not None:
                meta = style.meta
                if "offset" in meta:
                    offset_x, offset_y = style.meta["offset"]
                    offset_x2 = offset_x + len(segment.text)

                    if x < end and x >= start:
                        if x == end - 1:
                            segment_offset = len(segment.text)
                        else:
                            first, _ = segment.split_cells(x - start)
                            segment_offset = len(first.text)
                        return widget, (
                            None
                            if offset_y is None
                            else Offset(offset_x + segment_offset, offset_y)
                        )
            start = end

        return widget, (None if offset_y is None else Offset(offset_x2, offset_y))

    def find_widget(self, widget: Widget) -> MapGeometry:
        """Get information regarding the relative position of a widget in the Compositor.

        Args:
            widget: The Widget in this layout you wish to know the Region of.

        Raises:
            NoWidget: If the Widget is not contained in this Layout.

        Returns:
            Widget's composition information.
        """
        if self.root is None:
            raise errors.NoWidget("Widget is not in layout")
        try:
            if not self._full_map_invalidated:
                try:
                    return self._full_map[widget]
                except KeyError:
                    pass
            if self._visible_map is not None:
                try:
                    return self._visible_map[widget]
                except KeyError:
                    pass
            region = self.full_map[widget]
        except KeyError:
            raise errors.NoWidget("Widget is not in layout")
        else:
            return region

    @property
    def cuts(self) -> list[list[int]]:
        """Get vertical cuts.

        A cut is every point on a line where a widget starts or ends.

        Returns:
            A list of cuts for every line.
        """
        if self._cuts is not None:
            return self._cuts

        width, height = self.size
        cuts = [[0, width] for _ in range(height)]

        intersection = Region.intersection
        extend = list.extend

        for region, clip in self.visible_widgets.values():
            x, y, region_width, region_height = intersection(region, clip)
            if region_width and region_height:
                region_cuts = (x, x + region_width)
                for cut in cuts[y : y + region_height]:
                    extend(cut, region_cuts)

        # Sort the cuts for each line
        self._cuts = [sorted(set(line_cuts)) for line_cuts in cuts]

        return self._cuts

    def _get_renders(
        self, crop: Region | None = None
    ) -> Iterable[tuple[Region, Region, list[Strip]]]:
        """Get rendered widgets (lists of segments) in the composition.

        Args:
            crop: Region to crop to, or `None` for entire screen.

        Returns:
            An iterable of <region>, <clip region>, and <strips>
        """
        # If a renderable throws an error while rendering, the user likely doesn't care about the traceback
        # up to this point.
        _rich_traceback_guard = True

        _Region = Region

        visible_widgets = self.visible_widgets

        if crop:
            crop_overlaps = crop.overlaps
            widget_regions = [
                (widget, region, clip)
                for widget, (region, clip) in visible_widgets.items()
                if crop_overlaps(clip)
            ]
        else:
            widget_regions = [
                (widget, region, clip)
                for widget, (region, clip) in visible_widgets.items()
            ]

        intersection = _Region.intersection
        contains_region = _Region.contains_region

        for widget, region, clip in widget_regions:
            if contains_region(clip, region):
                yield (
                    region,
                    clip,
                    widget.render_lines(
                        _Region(
                            0,
                            0,
                            region.width,
                            region.height,
                        )
                    ),
                )
            else:
                new_x, new_y, new_width, new_height = intersection(region, clip)
                if new_width and new_height:
                    yield (
                        region,
                        clip,
                        widget.render_lines(
                            _Region(
                                new_x - region.x,
                                new_y - region.y,
                                new_width,
                                new_height,
                            )
                        ),
                    )

    def render_update(
        self,
        full: bool = False,
        screen_stack: list[Screen] | None = None,
        simplify: bool = False,
    ) -> RenderableType | None:
        """Render an update renderable.

        Args:
            full: Perform a full update if `True`, otherwise a partial update.
            screen_stack: Screen stack list. Defaults to None.
            simplify: Simplify segments.

        Returns:
            A renderable for the update, or `None` if no update was required.
        """

        visible_screen_stack.set([] if screen_stack is None else screen_stack)
        screen_region = self.size.region
        if full or screen_region in self._dirty_regions:
            return self.render_full_update(simplify=simplify)
        else:
            return self.render_partial_update()

    def render_inline(
        self,
        size: Size,
        screen_stack: list[Screen] | None = None,
        clear: bool = False,
    ) -> RenderableType:
        """Render an inline update.

        Args:
            size: Inline size.
            screen_stack: Screen stack list. Defaults to None.
            clear: Also clear below the inline update (set when size decreases).

        Returns:
            A renderable.
        """
        visible_screen_stack.set([] if screen_stack is None else screen_stack)
        strips = self.render_strips(size)
        return InlineUpdate(strips, clear=clear)

    def render_full_update(self, simplify: bool = False) -> LayoutUpdate:
        """Render a full update.

        Args:
            simplify: Simplify the segments (combine contiguous segments).

        Returns:
            A LayoutUpdate renderable.
        """
        screen_region = self.size.region
        self._dirty_regions.clear()
        crop = screen_region
        chops = self._render_chops(crop, lambda y: True)
        if simplify:
            render_strips = [
                Strip.join(chop.values()).simplify().discard_meta() for chop in chops
            ]
        else:
            render_strips = [Strip.join(chop.values()) for chop in chops]

        return LayoutUpdate(render_strips, screen_region)

    def render_partial_update(self) -> ChopsUpdate | None:
        """Render a partial update.

        Returns:
            A ChopsUpdate if there is anything to update, otherwise `None`.
        """
        screen_region = self.size.region
        update_regions = self._dirty_regions.copy()
        self._dirty_regions.clear()
        if update_regions:
            # Create a crop region that surrounds all updates.
            crop = Region.from_union(update_regions).intersection(screen_region)
            spans = list(self._regions_to_spans(update_regions))
            is_rendered_line = {y for y, _, _ in spans}.__contains__
        else:
            return None
        chops = self._render_chops(crop, is_rendered_line)
        chop_ends = [cut_set[1:] for cut_set in self.cuts]
        return ChopsUpdate(chops, spans, chop_ends)

    def render_strips(self, size: Size | None = None) -> list[Strip]:
        """Render to a list of strips.

        Args:
            size: Size of render.

        Returns:
            A list of strips with the screen content.
        """
        if size is None:
            size = self.size
        chops = self._render_chops(size.region, lambda y: True)
        render_strips = [Strip.join(chop.values()) for chop in chops[: size.height]]
        return render_strips

    def _render_chops(
        self,
        crop: Region,
        is_rendered_line: Callable[[int], bool],
    ) -> Sequence[Mapping[int, Strip | None]]:
        """Render update 'chops'.

        Args:
            crop: Region to crop to.
            is_rendered_line: Callable to check if line should be rendered.

        Returns:
            Chops structure.
        """
        cuts = self.cuts
        fromkeys = cast("Callable[[list[int]], dict[int, Strip | None]]", dict.fromkeys)
        chops: list[dict[int, Strip | None]]
        chops = [fromkeys(cut_set[:-1]) for cut_set in cuts]

        cut_strips: Iterable[Strip]

        # Go through all the renders in reverse order and fill buckets with no render
        renders = self._get_renders(crop)
        intersection = Region.intersection

        for region, clip, strips in renders:
            render_region = intersection(region, clip)
            render_x = render_region.x
            first_cut, last_cut = render_region.column_span

            for y, strip in zip(render_region.line_range, strips):
                if not is_rendered_line(y):
                    continue

                chops_line = chops[y]
                final_cuts = [cut for cut in cuts[y] if (last_cut >= cut >= first_cut)]
                cut_strips = strip.divide([cut - render_x for cut in final_cuts[1:]])

                # Since we are painting front to back, the first segments for a cut "wins"
                get_chops_line = chops_line.get
                for cut, strip in zip(final_cuts, cut_strips):
                    if get_chops_line(cut) is None:
                        chops_line[cut] = strip

        return chops

    def __rich__(self) -> StripRenderable:
        return StripRenderable(self.render_strips())

    def update_widgets(self, widgets: set[Widget]) -> None:
        """Update the given widgets in the composition.

        Args:
            widgets: Set of Widgets to update.
        """

        # If there are any *new* widgets we need to invalidate the full map
        if not self._full_map_invalidated and not widgets.issubset(
            self.visible_widgets.keys()
        ):
            self._full_map_invalidated = True

        regions: list[Region] = []
        add_region = regions.append
        get_widget = self.visible_widgets.__getitem__
        for widget in self.visible_widgets.keys() & widgets:
            region, clip = get_widget(widget)
            offset = region.offset
            intersection = clip.intersection
            for dirty_region in widget._exchange_repaint_regions():
                update_region = intersection(dirty_region.translate(offset))
                if update_region:
                    add_region(update_region)

        self._dirty_regions.update(regions)



from operator import attrgetter

from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import Label


class Number(Label):
    DEFAULT_CSS = """
    Number {
        width: 1fr;
    }
    """

    def __init__(self, number: int) -> None:
        self.number = number
        super().__init__(classes=f"number{number}")

    def render(self) -> str:
        return str(self.number)


class NumberList(Vertical):

    DEFAULT_CSS = """
    NumberList {
        width: 1fr;
        Number {
            border: green;
            box-sizing: border-box;
            &.number1 {
                height: 3;
            }
            &.number2 {
                height: 4;
            }
            &.number3 {
                height: 5;
            }
            &.number4 {
                height: 6;
            }
            &.number5 {
                height: 7;
            }
        }
    }

    """

    def compose(self) -> ComposeResult:
        yield Number(5)
        yield Number(1)
        yield Number(3)
        yield Number(2)
        yield Number(4)


class SortApp(App):

    def compose(self) -> ComposeResult:
        with Horizontal():
            yield NumberList(id="unsorted")
            yield NumberList(id="ascending")
            yield NumberList(id="descending")

    def on_mount(self) -> None:
        self.query_one("#ascending").sort_children(
            key=attrgetter("number"),
        )
        self.query_one("#descending").sort_children(
            key=attrgetter("number"),
            reverse=True,
        )


if __name__ == "__main__":
    app = SortApp()
    app.run()



from textual._loop import loop_first, loop_first_last, loop_from_index, loop_last


def test_loop_first():
    assert list(loop_first([])) == []
    iterable = loop_first(["apples", "oranges", "pears", "lemons"])
    assert next(iterable) == (True, "apples")
    assert next(iterable) == (False, "oranges")
    assert next(iterable) == (False, "pears")
    assert next(iterable) == (False, "lemons")


def test_loop_last():
    assert list(loop_last([])) == []
    iterable = loop_last(["apples", "oranges", "pears", "lemons"])
    assert next(iterable) == (False, "apples")
    assert next(iterable) == (False, "oranges")
    assert next(iterable) == (False, "pears")
    assert next(iterable) == (True, "lemons")


def test_loop_first_last():
    assert list(loop_first_last([])) == []
    iterable = loop_first_last(["apples", "oranges", "pears", "lemons"])
    assert next(iterable) == (True, False, "apples")
    assert next(iterable) == (False, False, "oranges")
    assert next(iterable) == (False, False, "pears")
    assert next(iterable) == (False, True, "lemons")


def test_loop_from_index():
    assert list(loop_from_index("abcdefghij", 3)) == [
        (4, "e"),
        (5, "f"),
        (6, "g"),
        (7, "h"),
        (8, "i"),
        (9, "j"),
        (0, "a"),
        (1, "b"),
        (2, "c"),
        (3, "d"),
    ]

    assert list(loop_from_index("abcdefghij", 3, direction=-1)) == [
        (2, "c"),
        (1, "b"),
        (0, "a"),
        (9, "j"),
        (8, "i"),
        (7, "h"),
        (6, "g"),
        (5, "f"),
        (4, "e"),
        (3, "d"),
    ]

    assert list(loop_from_index("abcdefghij", 3, wrap=False)) == [
        (4, "e"),
        (5, "f"),
        (6, "g"),
        (7, "h"),
        (8, "i"),
        (9, "j"),
    ]



from textual.app import App, ComposeResult
from textual.widgets import Label


class PositionApp(App):
    CSS_PATH = "position.tcss"

    def compose(self) -> ComposeResult:
        yield Label("Absolute", id="label1")
        yield Label("Relative", id="label2")


if __name__ == "__main__":
    app = PositionApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal
from textual.widgets import Button, Footer, Header, Static

QUESTION = "Do you want to learn about Textual CSS?"


class ExampleApp(App):
    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Container(
            Static(QUESTION, classes="question"),
            Horizontal(
                Button("Yes", variant="success"),
                Button("No", variant="error"),
                classes="buttons",
            ),
            id="dialog",
        )


if __name__ == "__main__":
    app = ExampleApp()
    app.run()



import pytest

from textual.color import Color
from textual.style import Style


@pytest.mark.parametrize(
    ["markup", "style"],
    [
        ("", Style()),
        (
            "b",
            Style(bold=True),
        ),
        ("i", Style(italic=True)),
        ("u", Style(underline=True)),
        ("r", Style(reverse=True)),
        ("bold", Style(bold=True)),
        ("italic", Style(italic=True)),
        ("underline", Style(underline=True)),
        ("reverse", Style(reverse=True)),
        ("bold italic", Style(bold=True, italic=True)),
        ("not bold italic", Style(bold=False, italic=True)),
        ("bold not italic", Style(bold=True, italic=False)),
        ("rgb(10, 20, 30)", Style(foreground=Color(10, 20, 30))),
        ("rgba(10, 20, 30, 0.5)", Style(foreground=Color(10, 20, 30, 0.5))),
        ("rgb(10, 20, 30) 50%", Style(foreground=Color(10, 20, 30, 0.5))),
        ("on rgb(10, 20, 30)", Style(background=Color(10, 20, 30))),
        ("on rgb(10, 20, 30) 50%", Style(background=Color(10, 20, 30, 0.5))),
        ("@click=app.bell", Style.from_meta({"@click": "app.bell"})),
        ("@click='app.bell'", Style.from_meta({"@click": "app.bell"})),
        ('''@click="app.bell"''', Style.from_meta({"@click": "app.bell"})),
        ("@click=app.bell()", Style.from_meta({"@click": "app.bell()"})),
        (
            "@click=app.notify('hello')",
            Style.from_meta({"@click": "app.notify('hello')"}),
        ),
        (
            "@click=app.notify('hello [World]!')",
            Style.from_meta({"@click": "app.notify('hello [World]!')"}),
        ),
        (
            "@click=app.notify('hello') bold",
            Style(bold=True) + Style.from_meta({"@click": "app.notify('hello')"}),
        ),
    ],
)
def test_parse_style(markup: str, style: Style) -> None:
    """Check parsing of valid styles."""
    parsed_style = Style.parse(markup)
    print("parsed\t\t", repr(parsed_style))
    print("expected\t", repr(style))
    print(parsed_style.meta, style.meta)
    print(parsed_style._meta)
    print(style._meta)
    assert parsed_style == style



from __future__ import annotations

from rich.segment import Segment
from rich.style import Style

from textual._styles_cache import StylesCache
from textual.color import Color
from textual.css.styles import Styles
from textual.geometry import Region, Size
from textual.strip import Strip


def _extract_content(lines: list[Strip]) -> list[str]:
    """Extract the text content from lines."""
    content = ["".join(segment.text for segment in line) for line in lines]
    return content


def test_set_dirty():
    cache = StylesCache()
    cache.set_dirty(Region(3, 4, 10, 2))
    assert not cache.is_dirty(3)
    assert cache.is_dirty(4)
    assert cache.is_dirty(5)
    assert not cache.is_dirty(6)


def test_no_styles():
    """Test that empty style returns the content un-altered"""
    content = [
        Strip([Segment("foo")]),
        Strip([Segment("bar")]),
        Strip([Segment("baz")]),
    ]
    styles = Styles()
    cache = StylesCache()
    lines = cache.render(
        styles,
        Size(3, 3),
        Color.parse("blue"),
        Color.parse("green"),
        content.__getitem__,
        None,
        None,
        content_size=Size(3, 3),
    )
    style = Style.from_color(bgcolor=Color.parse("green").rich_color)

    expected = [
        Strip([Segment("foo", style)], 3),
        Strip([Segment("bar", style)], 3),
        Strip([Segment("baz", style)], 3),
    ]

    print(lines[0])
    print(expected[0])

    assert lines == expected


def test_border():
    content = [
        Strip([Segment("foo")]),
        Strip([Segment("bar")]),
        Strip([Segment("baz")]),
    ]
    styles = Styles()
    styles.border = ("heavy", "white")
    cache = StylesCache()
    lines = cache.render(
        styles,
        Size(5, 5),
        Color.parse("blue"),
        Color.parse("green"),
        content.__getitem__,
        None,
        None,
        content_size=Size(3, 3),
    )

    text_content = _extract_content(lines)

    expected_text = [
        "┏━━━┓",
        "┃foo┃",
        "┃bar┃",
        "┃baz┃",
        "┗━━━┛",
    ]

    assert text_content == expected_text


def test_padding():
    content = [
        Strip([Segment("foo")]),
        Strip([Segment("bar")]),
        Strip([Segment("baz")]),
    ]
    styles = Styles()
    styles.padding = 1
    cache = StylesCache()
    lines = cache.render(
        styles,
        Size(5, 5),
        Color.parse("blue"),
        Color.parse("green"),
        content.__getitem__,
        None,
        None,
        content_size=Size(3, 3),
    )

    text_content = _extract_content(lines)

    expected_text = [
        "     ",
        " foo ",
        " bar ",
        " baz ",
        "     ",
    ]

    assert text_content == expected_text


def test_padding_border():
    content = [
        Strip([Segment("foo")]),
        Strip([Segment("bar")]),
        Strip([Segment("baz")]),
    ]
    styles = Styles()
    styles.padding = 1
    styles.border = ("heavy", "white")
    cache = StylesCache()
    lines = cache.render(
        styles,
        Size(7, 7),
        Color.parse("blue"),
        Color.parse("green"),
        content.__getitem__,
        None,
        None,
        content_size=Size(3, 3),
    )

    text_content = _extract_content(lines)

    expected_text = [
        "┏━━━━━┓",
        "┃     ┃",
        "┃ foo ┃",
        "┃ bar ┃",
        "┃ baz ┃",
        "┃     ┃",
        "┗━━━━━┛",
    ]

    assert text_content == expected_text


def test_outline():
    content = [
        Strip([Segment("foo")]),
        Strip([Segment("bar")]),
        Strip([Segment("baz")]),
    ]
    styles = Styles()
    styles.outline = ("heavy", "white")
    cache = StylesCache()
    lines = cache.render(
        styles,
        Size(3, 3),
        Color.parse("blue"),
        Color.parse("green"),
        content.__getitem__,
        None,
        None,
        content_size=Size(3, 3),
    )

    text_content = _extract_content(lines)
    expected_text = [
        "┏━┓",
        "┃a┃",
        "┗━┛",
    ]
    assert text_content == expected_text


def test_crop():
    content = [
        Strip([Segment("foo")]),
        Strip([Segment("bar")]),
        Strip([Segment("baz")]),
    ]
    styles = Styles()
    styles.padding = 1
    styles.border = ("heavy", "white")
    cache = StylesCache()
    lines = cache.render(
        styles,
        Size(7, 7),
        Color.parse("blue"),
        Color.parse("green"),
        content.__getitem__,
        None,
        None,
        content_size=Size(3, 3),
        crop=Region(2, 2, 3, 3),
    )
    text_content = _extract_content(lines)
    expected_text = [
        "foo",
        "bar",
        "baz",
    ]
    assert text_content == expected_text


def test_dirty_cache() -> None:
    """Check that we only render content once or if it has been marked as dirty."""

    content = [
        Strip([Segment("foo")]),
        Strip([Segment("bar")]),
        Strip([Segment("baz")]),
    ]
    rendered_lines: list[int] = []

    def get_content_line(y: int) -> Strip:
        rendered_lines.append(y)
        return content[y]

    styles = Styles()
    styles.padding = 1
    styles.border = ("heavy", "white")
    cache = StylesCache()
    lines = cache.render(
        styles,
        Size(7, 7),
        Color.parse("blue"),
        Color.parse("green"),
        get_content_line,
        None,
        None,
        content_size=Size(3, 3),
    )
    assert rendered_lines == [0, 1, 2]
    del rendered_lines[:]

    text_content = _extract_content(lines)

    expected_text = [
        "┏━━━━━┓",
        "┃     ┃",
        "┃ foo ┃",
        "┃ bar ┃",
        "┃ baz ┃",
        "┃     ┃",
        "┗━━━━━┛",
    ]
    assert text_content == expected_text

    # Re-render styles, check that content was not requested
    lines = cache.render(
        styles,
        Size(7, 7),
        Color.parse("blue"),
        Color.parse("green"),
        get_content_line,
        None,
        None,
        content_size=Size(3, 3),
    )
    assert rendered_lines == []
    del rendered_lines[:]
    text_content = _extract_content(lines)
    assert text_content == expected_text

    # Mark 2 lines as dirty
    cache.set_dirty(Region(0, 2, 7, 2))

    lines = cache.render(
        styles,
        Size(7, 7),
        Color.parse("blue"),
        Color.parse("green"),
        get_content_line,
        None,
        None,
        content_size=Size(3, 3),
    )
    assert rendered_lines == [0, 1]
    text_content = _extract_content(lines)
    assert text_content == expected_text



from textual.app import App, ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.widgets import Button, ContentSwitcher, DataTable, Markdown

MARKDOWN_EXAMPLE = """# Three Flavours Cornetto

The Three Flavours Cornetto trilogy is an anthology series of British
comedic genre films directed by Edgar Wright.

## Shaun of the Dead

| Flavour | UK Release Date | Director |
| -- | -- | -- |
| Strawberry | 2004-04-09 | Edgar Wright |

## Hot Fuzz

| Flavour | UK Release Date | Director |
| -- | -- | -- |
| Classico | 2007-02-17 | Edgar Wright |

## The World's End

| Flavour | UK Release Date | Director |
| -- | -- | -- |
| Mint | 2013-07-19 | Edgar Wright |
"""


class ContentSwitcherApp(App[None]):
    CSS_PATH = "content_switcher.tcss"

    def compose(self) -> ComposeResult:
        with Horizontal(id="buttons"):  # (1)!
            yield Button("DataTable", id="data-table")  # (2)!
            yield Button("Markdown", id="markdown")  # (3)!

        with ContentSwitcher(initial="data-table"):  # (4)!
            yield DataTable(id="data-table")
            with VerticalScroll(id="markdown"):
                yield Markdown(MARKDOWN_EXAMPLE)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.query_one(ContentSwitcher).current = event.button.id  # (5)!

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        table.add_columns("Book", "Year")
        table.add_rows(
            [
                (title.ljust(35), year)
                for title, year in (
                    ("Dune", 1965),
                    ("Dune Messiah", 1969),
                    ("Children of Dune", 1976),
                    ("God Emperor of Dune", 1981),
                    ("Heretics of Dune", 1984),
                    ("Chapterhouse: Dune", 1985),
                )
            ]
        )


if __name__ == "__main__":
    ContentSwitcherApp().run()



from textual.app import App, ComposeResult
from textual.widgets import Static


class BaseTester(Static, can_focus=True):
    DEFAULT_CSS = """
    BaseTester:focus {
        background: yellow;
        border: thick magenta;
    }
    """


class NonNestedCSS(BaseTester):
    DEFAULT_CSS = """
    NonNestedCSS {
        width: 1fr;
        height: 1fr;
        background: red 10%;
        border: blank;
    }

    NonNestedCSS:focus {
        background: red 20%;
        border: round red;
    }
    """


class NestedCSS(BaseTester):
    DEFAULT_CSS = """
    NestedCSS {
        width: 1fr;
        height: 1fr;        

        &:focus {
            background: green 20%;
            border: round green;
        }

        background: green 10%;
        border: blank;
    }
    """


class NestedPseudoClassesApp(App[None]):
    AUTO_FOCUS = "NestedCSS"

    CSS = """
    Screen {
        layout: horizontal;
    }
    """

    def compose(self) -> ComposeResult:
        yield NonNestedCSS("This isn't using nested CSS")
        yield NestedCSS("This is using nested CSS")


if __name__ == "__main__":
    NestedPseudoClassesApp().run()



from typing import TYPE_CHECKING, Any, Awaitable, Callable, List, Literal, Union

from typing_extensions import Protocol

if TYPE_CHECKING:
    from rich.segment import Segment

    from textual.message import Message


class MessageTarget(Protocol):
    """Protocol that must be followed by objects that can receive messages."""

    async def _post_message(self, message: "Message") -> bool: ...

    def post_message(self, message: "Message") -> bool: ...


class EventTarget(Protocol):
    async def _post_message(self, message: "Message") -> bool: ...

    def post_message(self, message: "Message") -> bool: ...


class UnusedParameter:
    """Helper type for a parameter that isn't specified in a method call."""


SegmentLines = List[List["Segment"]]
CallbackType = Union[Callable[[], Awaitable[None]], Callable[[], None]]
"""Type used for arbitrary callables used in callbacks."""
IgnoreReturnCallbackType = Union[Callable[[], Awaitable[Any]], Callable[[], Any]]
"""A callback which ignores the return type."""
WatchCallbackBothValuesType = Union[
    Callable[[Any, Any], Awaitable[None]],
    Callable[[Any, Any], None],
]
"""Type for watch methods that accept the old and new values of reactive objects."""
WatchCallbackNewValueType = Union[
    Callable[[Any], Awaitable[None]],
    Callable[[Any], None],
]
"""Type for watch methods that accept only the new value of reactive objects."""
WatchCallbackNoArgsType = Union[
    Callable[[], Awaitable[None]],
    Callable[[], None],
]
"""Type for watch methods that do not require the explicit value of the reactive."""
WatchCallbackType = Union[
    WatchCallbackBothValuesType,
    WatchCallbackNewValueType,
    WatchCallbackNoArgsType,
]
"""Type used for callbacks passed to the `watch` method of widgets."""

AnimationLevel = Literal["none", "basic", "full"]
"""The levels that the [`TEXTUAL_ANIMATIONS`][textual.constants.TEXTUAL_ANIMATIONS] env var can be set to."""



"""
Code browser example.

Run with:

    python code_browser.py PATH
"""

from __future__ import annotations

import sys

from rich.syntax import Syntax
from rich.traceback import Traceback

from textual.app import App, ComposeResult
from textual.containers import Container, VerticalScroll
from textual.reactive import reactive, var
from textual.widgets import DirectoryTree, Footer, Header, Static


class CodeBrowser(App):
    """Textual code browser app."""

    CSS_PATH = "code_browser.tcss"
    BINDINGS = [
        ("f", "toggle_files", "Toggle Files"),
        ("q", "quit", "Quit"),
    ]

    show_tree = var(True)
    path: reactive[str | None] = reactive(None)

    def watch_show_tree(self, show_tree: bool) -> None:
        """Called when show_tree is modified."""
        self.set_class(show_tree, "-show-tree")

    def compose(self) -> ComposeResult:
        """Compose our UI."""
        path = "./" if len(sys.argv) < 2 else sys.argv[1]
        yield Header()
        with Container():
            yield DirectoryTree(path, id="tree-view")
            with VerticalScroll(id="code-view"):
                yield Static(id="code", expand=True)
        yield Footer()

    def on_mount(self) -> None:
        self.query_one(DirectoryTree).focus()

        def theme_change(_signal) -> None:
            """Force the syntax to use a different theme."""
            self.watch_path(self.path)

        self.theme_changed_signal.subscribe(self, theme_change)

    def on_directory_tree_file_selected(
        self, event: DirectoryTree.FileSelected
    ) -> None:
        """Called when the user click a file in the directory tree."""
        event.stop()
        self.path = str(event.path)

    def watch_path(self, path: str | None) -> None:
        """Called when path changes."""
        code_view = self.query_one("#code", Static)
        if path is None:
            code_view.update("")
            return
        try:
            syntax = Syntax.from_path(
                path,
                line_numbers=True,
                word_wrap=False,
                indent_guides=True,
                theme="github-dark" if self.current_theme.dark else "github-light",
            )
        except Exception:
            code_view.update(Traceback(theme="github-dark", width=None))
            self.sub_title = "ERROR"
        else:
            code_view.update(syntax)
            self.query_one("#code-view").scroll_home(animate=False)
            self.sub_title = path

    def action_toggle_files(self) -> None:
        """Called in response to key binding."""
        self.show_tree = not self.show_tree


if __name__ == "__main__":
    CodeBrowser().run()



from textual.app import App, ComposeResult
from textual.color import Color
from textual.widgets import Static


class ColorApp(App):
    def compose(self) -> ComposeResult:
        self.widget1 = Static("Textual One")
        yield self.widget1
        self.widget2 = Static("Textual Two")
        yield self.widget2
        self.widget3 = Static("Textual Three")
        yield self.widget3

    def on_mount(self) -> None:
        self.widget1.styles.background = "#9932CC"
        self.widget2.styles.background = "hsl(150,42.9%,49.4%)"
        self.widget2.styles.color = "blue"
        self.widget3.styles.background = Color(191, 78, 96)


if __name__ == "__main__":
    app = ColorApp()
    app.run()



from textual.app import App
from textual.containers import Grid
from textual.widgets import Label


class ProgrammaticScrollbarGutterChange(App[None]):
    CSS = """
    Grid { grid-size: 2 2; scrollbar-size: 5 5; }
    Label { width: 100%; height: 100%; background: red; }
    """

    def compose(self):
        yield Grid(
            Label("one"),
            Label("two"),
            Label("three"),
            Label("four"),
        )

    def on_key(self, event):
        if event.key == "s":
            self.query_one(Grid).styles.scrollbar_gutter = "stable"


if __name__ == "__main__":
    app = ProgrammaticScrollbarGutterChange()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Markdown

TEST_CODE_MARKDOWN = """
# This is a H1
```python
def main():
    print("Hello world!")
```
"""


class MarkdownThemeSwitcherApp(App[None]):
    BINDINGS = [
        ("t", "toggle_theme"),
        ("d", "switch_dark"),
        ("l", "switch_light"),
    ]

    def action_switch_dark(self) -> None:
        md = self.query_one(Markdown)
        md.code_dark_theme = "solarized-dark"

    def action_switch_light(self) -> None:
        md = self.query_one(Markdown)
        md.code_light_theme = "solarized-light"

    def action_toggle_theme(self) -> None:
        self.theme = "textual-light" if self.theme == "textual-dark" else "textual-dark"

    def compose(self) -> ComposeResult:
        yield Markdown(TEST_CODE_MARKDOWN)


if __name__ == "__main__":
    MarkdownThemeSwitcherApp().run()



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Label


async def test_links() -> None:
    """Regression test for https://github.com/Textualize/textual/issues/4536"""
    messages: list[str] = []

    class LinkLabel(Label):
        def action_bell_message(self, message: str) -> None:
            nonlocal messages
            messages.append(f"label {message}")

    class HomeScreen(Screen[None]):
        def compose(self) -> ComposeResult:
            yield LinkLabel("[@click=app.bell_message('foo')]Ring the bell![/]")
            yield LinkLabel("[@click=screen.bell_message('bar')]Ring the bell![/]")
            yield LinkLabel("[@click=bell_message('baz')]Ring the bell![/]")

        def action_bell_message(self, message: str) -> None:
            nonlocal messages
            messages.append(f"screen {message}")

    class ScreenNamespace(App[None]):
        def get_default_screen(self) -> HomeScreen:
            return HomeScreen()

        def action_bell_message(self, message: str) -> None:
            nonlocal messages
            messages.append(f"app {message}")

    async with ScreenNamespace().run_test() as pilot:
        await pilot.click(offset=(5, 0))
        await pilot.click(offset=(5, 1))
        await pilot.click(offset=(5, 2))
        assert messages == ["app foo", "screen bar", "label baz"]



from textual.app import App, ComposeResult

from textual.widgets import Label

from textual.containers import VerticalGroup
from textual.content import Content

TEXT = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.
And when it has gone past, I will turn the inner eye to see its path.
Where the fear has gone there will be nothing. Only I will remain."""


class MyApp(App):
    CSS = """
    VerticalGroup {
        layout: grid;
        grid-size: 3 3;
        grid-columns: 1fr;
        grid-rows: auto;
        height: auto;      
    }
    Label {
        padding: 2 4;        
        border: heavy red;

	    &.left {
            text-align: left;
        }
        &.center {
            text-align: center;
        }
        &.right {
            text-align: right;
        }       
         &.justify {
            text-align: justify;
        }   
    }
    """

    def compose(self) -> ComposeResult:
        with VerticalGroup():
            yield Label(TEXT, classes="left", id="first")
            yield Label(TEXT, classes="left")
            yield Label(TEXT, classes="center")
            yield Label(TEXT, classes="center")
            yield Label(TEXT, classes="right")
            yield Label(TEXT, classes="right")
            yield Label(TEXT, classes="justify")
            yield Label(TEXT, classes="justify")
            yield Label(TEXT, classes="justify", id="last")


if __name__ == "__main__":
    MyApp().run()
